# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of teensy-core."""

def populate_download_cache_from_cipd(
    path: Path,
    download_path: Path,
    cipd_package_subpath: str,
) -> None:
    """Check for CIPD package availability."""
    package_path = path.parent.resolve()
    core_name = self.name
    core_cache_path = package_path / ".cache" / core_name
    core_cache_path.mkdir(parents=True, exist_ok=True)

    cipd_package_subpath = "pigweed_internal/third_party/"
    cipd_package_subpath += core_name
    cipd_package_subpath += "/${platform}"

    with tempfile.NamedTemporaryFile(
        prefix='cipd', delete=True, dir=core_cache_path
    ) as temp_json:
        temp_json_path = Path(temp_json.name)
        cipd_acl_check_command = [
            "cipd",
            "acl-check",
            cipd_package_subpath,
            "-reader",
            "-json-output",
            str(temp_json_path),
        ]
        subprocess.run(cipd_acl_check_command, capture_output=True)
        # Return if no packages are readable.
        if not temp_json_path.is_file():
            return
        result_text = temp_json.read_text()
        result_dict = json.loads(result_text)
        if 'result' not in result_dict:
            return



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for manipulating GN configs."""

from __future__ import annotations

from collections import deque
from json import loads as json_loads, dumps as json_dumps
from typing import Any, Deque, Dict, Iterable, Iterator, Optional, Set

from pw_build.gn_utils import GnLabel, MalformedGnError

GN_CONFIG_FLAGS = [
    'asmflags',
    'cflags',
    'cflags_c',
    'cflags_cc',
    'cflags_objc',
    'cflags_objcc',
    'defines',
    'include_dirs',
    'inputs',
    'ldflags',
    'lib_dirs',
    'libs',
    'precompiled_header',
    'precompiled_source',
    'rustenv',
    'rustflags',
    'swiftflags',
    'testonly',
]

_INTERNAL_FLAGS = ['nested', 'public_defines']


def _get_prefix(flag: str) -> str:
    """Returns the prefix used to identify values for a particular flag.

    Combining all values in a single set allows methods like `exact_cover` to
    analyze configs for patterns, but the values also need to be able to be
    separated out again. This prefix pattern is chosen as it is guaranteed not
    to be a valid source-relative path or label in GN or Bazel. It is
    encapsulated into its own function to encourage consistency and
    maintainability.

    Args:
        flag: The flag to convert to a prefix.
    """
    return f'{flag}::'


class GnConfig:
    """Represents a GN config.

    Attributes:
        label: The GN label of this config.
        values: A set of config values, prefixed by type. For example, a C
            compiler flag might be 'cflag::-foo'.
    """

    def __init__(
        self, public: bool = False, json: Optional[str] = None
    ) -> None:
        """Create a GN config object.

        Args:
            public: Indicates if this is a `public_config`.
            json: If provided, populates this object from a JSON string.
        """
        self.label: Optional[GnLabel] = None
        self.values: Set[str] = set()
        self._public: bool = public
        self._usages: int = 0
        if json:
            self._from_json(json)

    def __lt__(self, other: GnConfig) -> bool:
        """Compares two configs.

        A config is said to be "less than" another if it comes before the other
        when ordered according to the following rules, evaluated in order:
            * Public configs come before regular configs.
            * More frequently used configs come before those used less.
            * Shorter configs (in terms of values) come before longer ones.
            * Configs whose label comes before the other's comes first.
            * If all else fails, the configs are converted to strings and
              compared lexicographically.
        """
        if self._public != other.public():
            return not self._public and other.public()
        if self._usages != other._usages:
            return self._usages < other._usages
        if len(self.values) != len(other.values):
            return len(self.values) < len(other.values)
        if self.label != other.label:
            return str(self.label) < str(other.label)
        return str(self) < str(other)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, GnConfig)
            and self._public == other.public()
            and self._usages == other._usages
            and self.values == other.values
            and self.label == other.label
        )

    def __hash__(self) -> int:
        return hash((self._public, self._usages, str(self)))

    def __str__(self) -> str:
        return self.to_json()

    def __bool__(self) -> bool:
        return bool(self.values)

    def _from_json(self, data: str) -> None:
        """Populates this config from a JSON string.

        Args:
            data: A JSON representation of a config.
        """
        obj = json_loads(data)
        if 'label' in obj:
            self.label = GnLabel(obj['label'])
        for flag in GN_CONFIG_FLAGS + _INTERNAL_FLAGS:
            if flag in obj:
                self.add(flag, *obj[flag])
        if 'public' in obj:
            self._public = bool(obj['public'])
        if 'usages' in obj:
            self._usages = int(obj['usages'])

    def to_json(self) -> str:
        """Returns a JSON representation of this config."""
        obj: Dict[str, Any] = {}
        if self.label:
            obj['label'] = str(self.label)
        for flag in GN_CONFIG_FLAGS + _INTERNAL_FLAGS:
            if self.has(flag):
                obj[flag] = list(self.get(flag))
        if self._public:
            obj['public'] = self._public
        if self._usages:
            obj['usages'] = self._usages
        return json_dumps(obj)

    def has(self, flag: str) -> bool:
        """Returns whether this config has values for the given flag.

        Args:
            flag: The flag to check for.
        """
        return any(v.startswith(_get_prefix(flag)) for v in self.values)

    def add(self, flag: str, *values: str) -> None:
        """Adds a value to this config for the given flag.

        Args:
            flag: The flag to add values for.
        Variable Args:
            values: Strings to associate with the given flag.
        """
        if flag not in GN_CONFIG_FLAGS and flag not in _INTERNAL_FLAGS:
            raise MalformedGnError(f'invalid flag: {flag}')
        self.values |= {f'{_get_prefix(flag)}{v}' for v in values}

    def get(self, flag: str) -> Iterator[str]:
        """Iterates over the values for the given flag.

        Args:
            flag: the flag to look up.
        """
        prefix = _get_prefix(flag)
        for value in self.values:
            if value.startswith(prefix):
                yield value[len(prefix) :]

    def take(self, flag: str) -> Iterable[str]:
        """Extracts and returns the set of values for the given flag.

        Args:
            flag: The flag to remove and return values for.
        """
        prefix = _get_prefix(flag)
        taken = {v for v in self.values if v.startswith(prefix)}
        self.values = self.values - taken
        return [v[len(prefix) :] for v in taken]

    def deduplicate(self, *configs: GnConfig) -> Iterator[GnConfig]:
        """Removes values found in the given configs.

        Values are only removed if all of the values in a config are present.
        Returns the configs which resulte in values being removed.

        Variable Args:
            configs: The configs whose values should be removed from this config

        Raises:
            ValueError if any of the given configs do not have a label.
        """
        matching = []
        for config in configs:
            if not config.label:
                raise ValueError('config has no label')
            if not config:
                continue
            if config.values <= self.values:
                matching.append(config)
        matching.sort(key=lambda config: len(config.values), reverse=True)
        for config in matching:
            if config.values & self.values:
                self.values = self.values - config.values
                yield config

    def within(self, config: GnConfig) -> bool:
        """Returns whether the values of this config are a subset of another.

        Args:
            config: The config whose values are checked.
        """
        return self.values <= config.values

    def count_usages(self, configs: Iterable[GnConfig]) -> int:
        """Counts how many other configs this config is within.

        Args:
            configs: The set of configs which may contain this object's values.
        """
        self._usages = sum(map(self.within, configs))
        return self._usages

    def public(self) -> bool:
        """Returns whether this object represents a public config."""
        return self._public

    def extract_public(self) -> GnConfig:
        """Extracts and returns the set of values that need to be public.

        'include_dirs' and public 'defines' for a GN target need to be forwarded
        to anything that depends on that target.
        """
        public = GnConfig(public=True)
        public.add('include_dirs', *(self.take('include_dirs')))
        public.add('defines', *(self.take('public_defines')))
        return public

    def generate_label(self, label: GnLabel, index: int) -> None:
        """Creates a label for this config."""
        name = label.name().replace('-', '_')
        name = f'{name}_' or ''
        public = 'public_' if self._public else ''
        self.label = GnLabel(f'{label.dir()}:{name}{public}config{index}')


def _exact_cover(*configs: GnConfig) -> Iterator[GnConfig]:
    """Returns the exact covering set of configs for a given set of configs.

    An exact cover of a sequence of sets is the smallest set of subsets such
    that each element in the union of sets appears in exactly one subset. In
    other words, the subsets are disjoint and every set in the original sequence
    equals some union of subsets.

    As a side effect, this also separates public and regular flags, as GN
    targets have separate lists for `public_configs` and `configs`.

    Variables Args:
        configs: The set of configs to produce an exact cover for.
    """
    pending: Deque[Set[str]] = deque([config.values for config in configs])
    intermediate: Deque[Set[str]] = deque()
    disjoint: Deque[Set[str]] = deque()
    while pending:
        config_a = pending.popleft()
        if not config_a:
            continue
        ok = True
        while disjoint:
            intermediate.append(disjoint.popleft())
        while intermediate:
            config_b = intermediate.popleft()
            ok = False
            if config_a == config_b:
                disjoint.append(config_b)
                break
            if config_a < config_b:
                pending.append(config_b - config_a)
                disjoint.append(config_a)
                break
            if config_a > config_b:
                pending.append(config_a - config_b)
                disjoint.append(config_b)
                break
            config_c = config_a & config_b
            if config_c:
                pending.append(config_a - config_c)
                pending.append(config_b - config_c)
                pending.append(config_c)
                break
            ok = True
            disjoint.append(config_b)
        if ok:
            disjoint.append(config_a)

    for values in disjoint:
        config = GnConfig()
        config.values = values
        public = config.extract_public()
        if public:
            yield public
        if config:
            yield config


def _filter_by_usage(
    covers: Iterator[GnConfig], extract_public: bool, *configs: GnConfig
) -> Iterable[GnConfig]:
    """Filters configs to only include public or those used at least 3 times.

    Args:
        covers: A set of configs that is an exact cover for `configs`.
        extract_public: If true, all public configs are yielded.

    Variable Args:
        configs: A set of configs being conslidated.
    """
    for cover in covers:
        if cover.count_usages(configs) > 2 or (
            extract_public and cover.public()
        ):
            yield cover


def consolidate_configs(
    label: GnLabel, *configs: GnConfig, **kwargs
) -> Iterator[GnConfig]:
    """Extracts and returns the most common sub-configs across a set of configs.

    See also `_exact_cover`. An exact cover of configs can be used to find the
    most common sub-configs. These sub-configs are given labels, and then
    replaced in the original configs.

    Callers may optionally set the keyword argument of `extract_public` to
    `True` if all public configs should be extracted, regardless of usage count.
    Flags like `include_dirs` must be in a GN `public_config` to be forwarded,
    so this is useful for the first consolidation of configs corresponding to a
    Bazel package. Subsequent consolidations, i.e. for a group of Bazel
    packages, may want to avoid pulling public configs out of packages and omit
    this parameter.

    Args:
        label: The base label to use for generated configs.

    Variable Args:
        configs: Configs to examine for common, interesting shared sub-configs.

    Keyword Args:
        extract_public: If true, always considers public configs "interesting".
    """
    extract_public = kwargs.get('extract_public', False)
    covers = list(
        _filter_by_usage(_exact_cover(*configs), extract_public, *configs)
    )
    covers.sort(reverse=True)
    public_i = 1
    config_j = 1
    for cover in covers:
        if cover.public():
            cover.generate_label(label, public_i)
            public_i += 1
        else:
            cover.generate_label(label, config_j)
            config_j += 1
        yield cover



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_i2c/i2c.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_i2c/i2c.proto',
  package='pw.i2c',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x10pw_i2c/i2c.proto\x12\x06pw.i2c\"e\n\x0fI2cWriteRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\r\n\x05value\x18\x04 \x01(\x0c\"\x12\n\x10I2cWriteResponse\"h\n\x0eI2cReadRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\x11\n\tread_size\x18\x04 \x01(\r\" \n\x0fI2cReadResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\x32\x84\x01\n\x03I2c\x12?\n\x08I2cWrite\x12\x17.pw.i2c.I2cWriteRequest\x1a\x18.pw.i2c.I2cWriteResponse\"\x00\x12<\n\x07I2cRead\x12\x16.pw.i2c.I2cReadRequest\x1a\x17.pw.i2c.I2cReadResponse\"\x00\x62\x06proto3'
)




_I2CWRITEREQUEST = _descriptor.Descriptor(
  name='I2cWriteRequest',
  full_name='pw.i2c.I2cWriteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bus_index', full_name='pw.i2c.I2cWriteRequest.bus_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_address', full_name='pw.i2c.I2cWriteRequest.target_address', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='register_address', full_name='pw.i2c.I2cWriteRequest.register_address', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.i2c.I2cWriteRequest.value', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=129,
)


_I2CWRITERESPONSE = _descriptor.Descriptor(
  name='I2cWriteResponse',
  full_name='pw.i2c.I2cWriteResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=131,
  serialized_end=149,
)


_I2CREADREQUEST = _descriptor.Descriptor(
  name='I2cReadRequest',
  full_name='pw.i2c.I2cReadRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bus_index', full_name='pw.i2c.I2cReadRequest.bus_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_address', full_name='pw.i2c.I2cReadRequest.target_address', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='register_address', full_name='pw.i2c.I2cReadRequest.register_address', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_size', full_name='pw.i2c.I2cReadRequest.read_size', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=151,
  serialized_end=255,
)


_I2CREADRESPONSE = _descriptor.Descriptor(
  name='I2cReadResponse',
  full_name='pw.i2c.I2cReadResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.i2c.I2cReadResponse.value', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=289,
)

DESCRIPTOR.message_types_by_name['I2cWriteRequest'] = _I2CWRITEREQUEST
DESCRIPTOR.message_types_by_name['I2cWriteResponse'] = _I2CWRITERESPONSE
DESCRIPTOR.message_types_by_name['I2cReadRequest'] = _I2CREADREQUEST
DESCRIPTOR.message_types_by_name['I2cReadResponse'] = _I2CREADRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

I2cWriteRequest = _reflection.GeneratedProtocolMessageType('I2cWriteRequest', (_message.Message,), {
  'DESCRIPTOR' : _I2CWRITEREQUEST,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cWriteRequest)
  })
_sym_db.RegisterMessage(I2cWriteRequest)

I2cWriteResponse = _reflection.GeneratedProtocolMessageType('I2cWriteResponse', (_message.Message,), {
  'DESCRIPTOR' : _I2CWRITERESPONSE,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cWriteResponse)
  })
_sym_db.RegisterMessage(I2cWriteResponse)

I2cReadRequest = _reflection.GeneratedProtocolMessageType('I2cReadRequest', (_message.Message,), {
  'DESCRIPTOR' : _I2CREADREQUEST,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cReadRequest)
  })
_sym_db.RegisterMessage(I2cReadRequest)

I2cReadResponse = _reflection.GeneratedProtocolMessageType('I2cReadResponse', (_message.Message,), {
  'DESCRIPTOR' : _I2CREADRESPONSE,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cReadResponse)
  })
_sym_db.RegisterMessage(I2cReadResponse)



_I2C = _descriptor.ServiceDescriptor(
  name='I2c',
  full_name='pw.i2c.I2c',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=292,
  serialized_end=424,
  methods=[
  _descriptor.MethodDescriptor(
    name='I2cWrite',
    full_name='pw.i2c.I2c.I2cWrite',
    index=0,
    containing_service=None,
    input_type=_I2CWRITEREQUEST,
    output_type=_I2CWRITERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='I2cRead',
    full_name='pw.i2c.I2c.I2cRead',
    index=1,
    containing_service=None,
    input_type=_I2CREADREQUEST,
    output_type=_I2CREADRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_I2C)

DESCRIPTOR.services_by_name['I2c'] = _I2C

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_digital_io/digital_io.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_digital_io/digital_io.proto',
  package='pw.digital_io',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1epw_digital_io/digital_io.proto\x12\rpw.digital_io\"<\n\x16\x44igitalIoEnableRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12\x0e\n\x06\x65nable\x18\x02 \x01(\x08\"\x19\n\x17\x44igitalIoEnableResponse\"\\\n\x18\x44igitalIoSetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12,\n\x05state\x18\x02 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState\"\x1b\n\x19\x44igitalIoSetStateResponse\".\n\x18\x44igitalIoGetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\"I\n\x19\x44igitalIoGetStateResponse\x12,\n\x05state\x18\x01 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState**\n\x0e\x44igitalIoState\x12\x0c\n\x08INACTIVE\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x32\xa8\x02\n\tDigitalIo\x12Y\n\x06\x45nable\x12%.pw.digital_io.DigitalIoEnableRequest\x1a&.pw.digital_io.DigitalIoEnableResponse\"\x00\x12_\n\x08SetState\x12\'.pw.digital_io.DigitalIoSetStateRequest\x1a(.pw.digital_io.DigitalIoSetStateResponse\"\x00\x12_\n\x08GetState\x12\'.pw.digital_io.DigitalIoGetStateRequest\x1a(.pw.digital_io.DigitalIoGetStateResponse\"\x00\x62\x06proto3'
)

_DIGITALIOSTATE = _descriptor.EnumDescriptor(
  name='DigitalIoState',
  full_name='pw.digital_io.DigitalIoState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=384,
  serialized_end=426,
)
_sym_db.RegisterEnumDescriptor(_DIGITALIOSTATE)

DigitalIoState = enum_type_wrapper.EnumTypeWrapper(_DIGITALIOSTATE)
INACTIVE = 0
ACTIVE = 1



_DIGITALIOENABLEREQUEST = _descriptor.Descriptor(
  name='DigitalIoEnableRequest',
  full_name='pw.digital_io.DigitalIoEnableRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line_index', full_name='pw.digital_io.DigitalIoEnableRequest.line_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='enable', full_name='pw.digital_io.DigitalIoEnableRequest.enable', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=109,
)


_DIGITALIOENABLERESPONSE = _descriptor.Descriptor(
  name='DigitalIoEnableResponse',
  full_name='pw.digital_io.DigitalIoEnableResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=111,
  serialized_end=136,
)


_DIGITALIOSETSTATEREQUEST = _descriptor.Descriptor(
  name='DigitalIoSetStateRequest',
  full_name='pw.digital_io.DigitalIoSetStateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line_index', full_name='pw.digital_io.DigitalIoSetStateRequest.line_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='pw.digital_io.DigitalIoSetStateRequest.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=138,
  serialized_end=230,
)


_DIGITALIOSETSTATERESPONSE = _descriptor.Descriptor(
  name='DigitalIoSetStateResponse',
  full_name='pw.digital_io.DigitalIoSetStateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=232,
  serialized_end=259,
)


_DIGITALIOGETSTATEREQUEST = _descriptor.Descriptor(
  name='DigitalIoGetStateRequest',
  full_name='pw.digital_io.DigitalIoGetStateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line_index', full_name='pw.digital_io.DigitalIoGetStateRequest.line_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=261,
  serialized_end=307,
)


_DIGITALIOGETSTATERESPONSE = _descriptor.Descriptor(
  name='DigitalIoGetStateResponse',
  full_name='pw.digital_io.DigitalIoGetStateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='pw.digital_io.DigitalIoGetStateResponse.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=309,
  serialized_end=382,
)

_DIGITALIOSETSTATEREQUEST.fields_by_name['state'].enum_type = _DIGITALIOSTATE
_DIGITALIOGETSTATERESPONSE.fields_by_name['state'].enum_type = _DIGITALIOSTATE
DESCRIPTOR.message_types_by_name['DigitalIoEnableRequest'] = _DIGITALIOENABLEREQUEST
DESCRIPTOR.message_types_by_name['DigitalIoEnableResponse'] = _DIGITALIOENABLERESPONSE
DESCRIPTOR.message_types_by_name['DigitalIoSetStateRequest'] = _DIGITALIOSETSTATEREQUEST
DESCRIPTOR.message_types_by_name['DigitalIoSetStateResponse'] = _DIGITALIOSETSTATERESPONSE
DESCRIPTOR.message_types_by_name['DigitalIoGetStateRequest'] = _DIGITALIOGETSTATEREQUEST
DESCRIPTOR.message_types_by_name['DigitalIoGetStateResponse'] = _DIGITALIOGETSTATERESPONSE
DESCRIPTOR.enum_types_by_name['DigitalIoState'] = _DIGITALIOSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DigitalIoEnableRequest = _reflection.GeneratedProtocolMessageType('DigitalIoEnableRequest', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOENABLEREQUEST,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoEnableRequest)
  })
_sym_db.RegisterMessage(DigitalIoEnableRequest)

DigitalIoEnableResponse = _reflection.GeneratedProtocolMessageType('DigitalIoEnableResponse', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOENABLERESPONSE,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoEnableResponse)
  })
_sym_db.RegisterMessage(DigitalIoEnableResponse)

DigitalIoSetStateRequest = _reflection.GeneratedProtocolMessageType('DigitalIoSetStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOSETSTATEREQUEST,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoSetStateRequest)
  })
_sym_db.RegisterMessage(DigitalIoSetStateRequest)

DigitalIoSetStateResponse = _reflection.GeneratedProtocolMessageType('DigitalIoSetStateResponse', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOSETSTATERESPONSE,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoSetStateResponse)
  })
_sym_db.RegisterMessage(DigitalIoSetStateResponse)

DigitalIoGetStateRequest = _reflection.GeneratedProtocolMessageType('DigitalIoGetStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOGETSTATEREQUEST,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoGetStateRequest)
  })
_sym_db.RegisterMessage(DigitalIoGetStateRequest)

DigitalIoGetStateResponse = _reflection.GeneratedProtocolMessageType('DigitalIoGetStateResponse', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOGETSTATERESPONSE,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoGetStateResponse)
  })
_sym_db.RegisterMessage(DigitalIoGetStateResponse)



_DIGITALIO = _descriptor.ServiceDescriptor(
  name='DigitalIo',
  full_name='pw.digital_io.DigitalIo',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=429,
  serialized_end=725,
  methods=[
  _descriptor.MethodDescriptor(
    name='Enable',
    full_name='pw.digital_io.DigitalIo.Enable',
    index=0,
    containing_service=None,
    input_type=_DIGITALIOENABLEREQUEST,
    output_type=_DIGITALIOENABLERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetState',
    full_name='pw.digital_io.DigitalIo.SetState',
    index=1,
    containing_service=None,
    input_type=_DIGITALIOSETSTATEREQUEST,
    output_type=_DIGITALIOSETSTATERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetState',
    full_name='pw.digital_io.DigitalIo.GetState',
    index=2,
    containing_service=None,
    input_type=_DIGITALIOGETSTATEREQUEST,
    output_type=_DIGITALIOGETSTATERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DIGITALIO)

DESCRIPTOR.services_by_name['DigitalIo'] = _DIGITALIO

# @@protoc_insertion_point(module_scope)



import asyncio
import functools
import json
from pathlib import Path
from threading import Thread
import time
import websockets

from pw_console.console_log_server import (
    ConsoleLogHTTPRequestHandler,
    pw_console_http_server,
)

_PW_CONSOLE_MODULE = 'pw_console'

html_files = {
    '/{}'.format(t.name): t.read_text()
    for t in Path(__file__).parent.iterdir()
    if t.suffix in ['.css', '.html', '.js', '.json']
}


def _http_server_entry() -> None:
    handler = functools.partial(ConsoleLogHTTPRequestHandler, html_files)
    pw_console_http_server(3000, handler)


sample_logfiles = sorted(list(Path(__file__).parent.glob('*logs.json')))

websocket_server = None
websocket_port = None
websocket_loop = asyncio.new_event_loop()


async def _send_logs_over_websockets(websocket, _path) -> None:
    log_file = sample_logfiles[0].read_text()

    log_lines = list(log_file.splitlines())

    start_time = time.time()
    index = 0

    beginning_log_time = float(json.loads(log_lines[index])['time'])
    current_log_line = json.loads(log_lines[index])
    current_log_time = float(current_log_line['time'])

    while True:
        delta_time = time.time() - start_time
        if delta_time > (current_log_time - beginning_log_time):
            await websocket.send(log_lines[index])
            index += 1
            if index >= len(log_lines):
                # Restart from the beginning
                index = 0
                beginning_log_time = float(json.loads(log_lines[index])['time'])
            current_log_line = json.loads(log_lines[index])
            current_log_time = float(current_log_line['time'])
        else:
            await asyncio.sleep(0.1)


def _websocket_thread_entry():
    asyncio.set_event_loop(websocket_loop)
    websocket_server = websockets.serve(  # type: ignore # pylint: disable=no-member
        _send_logs_over_websockets, '127.0.0.1'
    )
    websocket_loop.run_until_complete(websocket_server)
    websocket_port = websocket_server.ws_server.sockets[0].getsockname()[1]
    websocket_running = True
    print(f'Serving on:\nhttp://127.0.0.1:3000/#ws={websocket_port}')

    websocket_loop.run_forever()


def main() -> None:
    if not sample_logfiles:
        print('ERROR: No files exist named *logs.json')

    print(f'Replaying: {sample_logfiles[0]}')

    server_thread = Thread(target=_http_server_entry, args=(), daemon=True)
    server_thread.start()

    websocket_thread = Thread(
        target=_websocket_thread_entry, args=(), daemon=True
    )
    websocket_thread.start()
    s = input('Press enter to quit.\n')


if __name__ == '__main__':
    main()



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Sphinx directives for Pigweed module metadata"""

from dataclasses import dataclass
from typing import cast, Dict, List, Optional, TypeVar, Union

# We use BeautifulSoup for certain docs rendering features. It may not be
# available in downstream projects. If so, no problem. We fall back to simpler
# docs rendering.
# pylint: disable=import-error
try:
    from bs4 import BeautifulSoup  # type: ignore
    from bs4.element import Tag as HTMLTag  # type: ignore

    bs_enabled = True
except ModuleNotFoundError:
    bs_enabled = False
# pylint: enable=import-error

import docutils
from docutils import nodes
import docutils.statemachine

# pylint: disable=consider-using-from-import
import docutils.parsers.rst.directives as directives  # type: ignore

# pylint: enable=consider-using-from-import
from sphinx.application import Sphinx as SphinxApplication
from sphinx.environment import BuildEnvironment
from sphinx.util.docutils import SphinxDirective

from sphinx_design.cards import CardDirective

EnvAttrT = TypeVar('EnvAttrT')


@dataclass
class ParsedBody:
    topnav: str
    body_without_topnav: str


class EnvMetadata:
    """Easier access to the Sphinx `env` for custom metadata.

    You can store things in the Sphinx `env`, which is just a dict. But each
    time you do, you have to handle the possibility that the key you want
    hasn't been set yet, and set it to a default. The `env` is also untyped,
    so you have to cast the value you get to whatever type you expect it to be.

    Or you can use this class to define your metadata keys up front, and just
    access them like: `value = EnvMetadata(env).my_value`

    ... which will handle initializing the value if it hasn't been yet and
    provide you a typed result.
    """

    def __init__(self, env: BuildEnvironment):
        self._env = env

    def _get_env_attr(self, attr: str, default: EnvAttrT) -> EnvAttrT:
        if not hasattr(self._env, attr):
            value: EnvAttrT = default
            setattr(self._env, attr, value)
        else:
            value = getattr(self._env, attr)

        return value

    @property
    def pw_parsed_bodies(self) -> Dict[str, ParsedBody]:
        default: Dict[str, ParsedBody] = {}
        return self._get_env_attr('pw_module_nav', default)


def status_choice(arg):
    return directives.choice(arg, ('experimental', 'unstable', 'stable'))


def status_badge(module_status: str) -> str:
    """Given a module status, return the status badge for rendering."""
    role = ':bdg-primary:'
    return role + f'`{module_status.title()}`'


def cs_url(module_name: str) -> str:
    """Return the codesearch URL for the given module."""
    return f'https://cs.opensource.google/pigweed/pigweed/+/main:{module_name}/'


def concat_tags(*tag_lists: List[str]) -> List[str]:
    """Given a list of tag lists, return them concat'ed and ready for render."""

    all_tags = tag_lists[0]

    for tag_list in tag_lists[1:]:
        if len(tag_list) > 0:
            all_tags.append(':octicon:`dot-fill`')
            all_tags.extend(tag_list)

    return all_tags


def create_topnav(
    subtitle: str,
    extra_classes: Optional[List[str]] = None,
) -> nodes.Node:
    """Create the nodes for the top title and navigation bar."""

    topnav_classes = (
        ['pw-topnav'] + extra_classes if extra_classes is not None else []
    )

    topnav_container = nodes.container(classes=topnav_classes)

    subtitle_node = nodes.paragraph(
        classes=['pw-topnav-subtitle'],
        text=subtitle,
    )

    topnav_container += subtitle_node
    return topnav_container


class PigweedModuleDirective(SphinxDirective):
    """Directive registering module metadata, rendering title & info card."""

    required_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        'name': directives.unchanged_required,
        'tagline': directives.unchanged_required,
        'status': status_choice,
        'is-deprecated': directives.flag,
        'languages': directives.unchanged,
        'code-size-impact': directives.unchanged,
        'facade': directives.unchanged,
        'nav': directives.unchanged_required,
    }

    def _try_get_option(self, option: str):
        """Try to get an option by name and raise on failure."""

        try:
            return self.options[option]
        except KeyError:
            raise self.error(f' :{option}: option is required')

    def _maybe_get_option(self, option: str):
        """Try to get an option by name and return None on failure."""
        return self.options.get(option, None)

    def run(self) -> List[nodes.Node]:
        module_name = self._try_get_option('name')
        tagline = self._try_get_option('tagline')

        status_tags: List[str] = [
            status_badge(self._try_get_option('status')),
        ]

        if 'is-deprecated' in self.options:
            status_tags.append(':bdg-danger:`Deprecated`')

        language_tags = []

        if 'languages' in self.options:
            languages = self.options['languages'].split(',')

            if len(languages) > 0:
                for language in languages:
                    language = language.strip()
                    if language == 'Rust':
                        language_tags.append(
                            f':bdg-link-info:`{language}'
                            + f'</rustdoc/{module_name}>`'
                        )
                    else:
                        language_tags.append(f':bdg-info:`{language}`')

        code_size_impact = []

        if code_size_text := self._maybe_get_option('code-size-impact'):
            code_size_impact.append(f'**Code Size Impact:** {code_size_text}')

        # Move the directive content into a section that we can render wherever
        # we want.
        raw_content = cast(List[str], self.content)  # type: ignore
        content = nodes.paragraph()
        self.state.nested_parse(raw_content, 0, content)

        # The card inherits its content from this node's content, which we've
        # already pulled out. So we can replace this node's content with the
        # content we need in the card.
        self.content = docutils.statemachine.StringList(
            concat_tags(status_tags, language_tags, code_size_impact)
        )

        card = CardDirective.create_card(
            inst=self,
            arguments=[],
            options={},
        )

        topbar = create_topnav(
            tagline,
            ['pw-module-index'],
        )

        return [topbar, card, content]


class PigweedModuleSubpageDirective(PigweedModuleDirective):
    """Directive registering module metadata, rendering title & info card."""

    required_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        'name': directives.unchanged_required,
        'tagline': directives.unchanged_required,
        'nav': directives.unchanged_required,
    }

    def run(self) -> List[nodes.Node]:
        tagline = self._try_get_option('tagline')

        topbar = create_topnav(
            tagline,
            ['pw-module-subpage'],
        )

        return [topbar]


def _parse_body(body: str) -> ParsedBody:
    """From the `body` HTML, return the topnav and the body without topnav.

    The fundamental idea is this: Our Sphinx directives can only render nodes
    *within* the docutils doc, but we want to elevate the top navbar *outside*
    of that doc into the web theme. Sphinx by itself provides no mechanism for
    this, since it's model looks something like this:

      ┌──────────────────┐
      │ Theme            │
      │  ┌──────────────┐│    When Sphinx builds HTML, the output is plain HTML
      │  │ Sphinx HTML  ││    with a structure defined by docutils. Themes can
      │  │              ││    build *around* that and cascade styles down *into*
      │  │              ││    that HTML, but there's no mechanism in the Sphinx
      │  └──────────────┘│    build to render docutils nodes in the theme.
      └──────────────────┘

    The escape hatch is this:
    - Render things within the Sphinx HTML output (`body`)
    - Use Sphinx theme templates to run code during the final render phase
    - Extract the HTML from the `body` and insert it in the theme via templates

    So this function extracts the things that we rendered in the `body` but
    actually want in the theme (the top navbar), returns them for rendering in
    the template, and returns the `body` with those things removed.
    """
    if not bs_enabled:
        return ParsedBody('', body)

    def _add_class_to_tag(tag: HTMLTag, classname: str) -> None:
        tag['class'] = tag.get('class', []) + [classname]  # type: ignore

    def _add_classes_to_tag(
        tag: HTMLTag, classnames: Union[str, List[str], None]
    ) -> None:
        tag['class'] = tag.get('class', []) + classnames  # type: ignore

    html = BeautifulSoup(body, features='html.parser')

    # Render the doc unchanged, unless it has the module doc topnav
    if (topnav := html.find('div', attrs={'class': 'pw-topnav'})) is None:
        return ParsedBody('', body)

    assert isinstance(topnav, HTMLTag)

    # Find the topnav title and subtitle
    topnav_title = topnav.find('p', attrs={'class': 'pw-topnav-title'})
    topnav_subtitle = topnav.find('p', attrs={'class': 'pw-topnav-subtitle'})
    assert isinstance(topnav_title, HTMLTag)
    assert isinstance(topnav_subtitle, HTMLTag)

    # Find the single `h1` element, the doc's canonical title
    doc_title = html.find('h1')
    assert isinstance(doc_title, HTMLTag)

    topnav_str = ''

    if 'pw-module-index' in topnav['class']:
        # Take the standard Sphinx/docutils title and add topnav styling
        _add_class_to_tag(doc_title, 'pw-topnav-title')
        # Replace the placeholder title in the topnav with the "official" `h1`
        topnav_title.replace_with(doc_title)
        # Promote the subtitle to `h2`
        topnav_subtitle.name = 'h2'
        # We're done mutating topnav; write it to string for rendering elsewhere
        topnav_str = str(topnav)
        # Destroy the instance that was rendered in the document
        topnav.decompose()

    elif 'pw-module-subpage' in topnav['class']:
        # Take the title from the topnav (the module name), promote it to `h1`
        topnav_title.name = 'h1'
        # Add the heading link, but pointed to the module index page
        heading_link = html.new_tag(
            'a',
            attrs={
                'class': ['headerlink'],
                'href': 'docs.html',
                'title': 'Permalink to module index',
            },
        )
        heading_link.string = '#'
        topnav_title.append(heading_link)
        # Promote the subtitle to `h2`
        topnav_subtitle.name = 'h2'
        # We're done mutating topnav; write it to string for rendering elsewhere
        topnav_str = str(topnav)
        # Destroy the instance that was rendered in the document
        topnav.decompose()

    return ParsedBody(topnav_str, str(html))


def setup_parse_body(_app, _pagename, _templatename, context, _doctree):
    def parse_body(body: str) -> ParsedBody:
        return _parse_body(body)

    context['parse_body'] = parse_body


def setup(app: SphinxApplication):
    app.add_directive('pigweed-module', PigweedModuleDirective)
    app.add_directive('pigweed-module-subpage', PigweedModuleSubpageDirective)

    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Configure Visual Studio Code (VSC) for Pigweed projects.

VSC recognizes three sources of configurable settings:

1. Project settings, stored in {project root}/.vscode/settings.json
2. Workspace settings, stored in (workspace root)/.vscode/settings.json;
   a workspace is a collection of projects/repositories that are worked on
   together in a single VSC instance
3. The user's personal settings, which are stored somewhere in the user's home
   directory, and are applied to all instances of VSC

This provides three levels of settings cascading:

    Workspace <- Project <- User

... where the arrow indicates the ability to override.

Out of these three, only project settings are useful to Pigweed projects. User
settings are essentially global and outside the scope of Pigweed. Workspaces are
seldom used and don't work well with the Pigweed directory structure.

Nonetheless, we want a three-tiered settings structure for Pigweed projects too:

A. Default settings provided by Pigweed, configuring VSC to use IDE features
B. Project-level overrides that downstream projects may define
C. User-level overrides that individual users may define

We accomplish all of that with only the project settings described in #1 above.

Default settings are defined in this module. Project settings can be defined in
.vscode/pw_project_settings.json and should be checked into the repository. User
settings can be defined in .vscode/pw_user_settings.json and should not be
checked into the repository. None of these settings have any effect until they
are merged into VSC's settings (.vscode/settings.json) via the functions in this
module. Those resulting settings are system-specific and should also not be
checked into the repository.

We provide the same structure to both tasks and extensions as well. Defaults
are provided by Pigweed, can be augmented or overridden at the project level
with .vscode/pw_project_tasks.json and .vscode/pw_project_extensions.json,
can be augmented or overridden by an individual developer with
.vscode/pw_user_tasks.json and .vscode/pw_user.extensions.json, and none of
this takes effect until they are merged into VSC's active settings files
(.vscode/tasks.json and .vscode/extensions.json) by running the appropriate
command.
"""

# TODO(chadnorvell): Import collections.OrderedDict when we don't need to
# support Python 3.8 anymore.
from enum import Enum
import json
import os
from pathlib import Path
import platform
import shutil
import subprocess
from typing import Any, Dict, List, Optional, OrderedDict

from pw_cli.env import pigweed_environment

from pw_ide.activate import BashShellModifier
from pw_ide.cpp import ClangdSettings, CppIdeFeaturesState

from pw_ide.editors import (
    EditorSettingsDict,
    EditorSettingsManager,
    EditorSettingsTypesWithDefaults,
    Json5FileFormat,
)

from pw_ide.python import PythonPaths
from pw_ide.settings import PigweedIdeSettings
from pw_ide.status_reporter import StatusReporter

env = pigweed_environment()

_VSIX_DIR = Path(env.PW_ROOT) / 'pw_ide' / 'vscode'


def _vsc_os(system: str = platform.system()):
    """Return the OS tag that VSC expects."""
    if system == 'Darwin':
        return 'osx'

    return system.lower()


def _activated_env() -> OrderedDict[str, Any]:
    """Return the environment diff needed to provide Pigweed activation.

    The integrated terminal will already include the user's default environment
    (e.g. from their shell init scripts). This provides the modifications to
    the environment needed for Pigweed activation.
    """
    # Not all environments have an actions.json, which this ultimately relies
    # on (e.g. tests in CI). No problem, just return an empty dict instead.
    try:
        activated_env = (
            BashShellModifier(env_only=True, path_var='${env:PATH}')
            .modify_env()
            .env_mod
        )
    except (FileNotFoundError, json.JSONDecodeError):
        activated_env = dict()

    return OrderedDict(activated_env)


def _local_terminal_integrated_env() -> Dict[str, Any]:
    """VSC setting to activate the integrated terminal."""
    return {f'terminal.integrated.env.{_vsc_os()}': _activated_env()}


def _local_clangd_settings(ide_settings: PigweedIdeSettings) -> Dict[str, Any]:
    """VSC settings for running clangd with Pigweed paths."""
    clangd_settings = ClangdSettings(ide_settings)
    return {
        'clangd.path': str(clangd_settings.clangd_path),
        'clangd.arguments': clangd_settings.arguments,
    }


def _local_python_settings() -> Dict[str, Any]:
    """VSC settings for finding the Python virtualenv."""
    paths = PythonPaths()
    return {
        'python.defaultInterpreterPath': str(paths.interpreter),
        'python.formatting.yapfPath': str(paths.bin_dir / 'yapf'),
    }


# The order is preserved despite starting with a plain dict because in Python
# 3.6+, plain dicts are actually ordered as an implementation detail. This could
# break on interpreters other than CPython, or if the implementation changes in
# the future. However, for now, this is much more readable than the more robust
# alternative of instantiating with a list of tuples.
_DEFAULT_SETTINGS: EditorSettingsDict = OrderedDict(
    {
        "editor.detectIndentation": False,
        "editor.rulers": [80],
        "editor.tabSize": 2,
        "files.associations": OrderedDict({"*.inc": "cpp"}),
        "files.exclude": OrderedDict(
            {
                "**/*.egg-info": True,
                "**/.mypy_cache": True,
                "**/__pycache__": True,
                ".cache": True,
                ".cipd": True,
                ".environment": True,
                ".presubmit": True,
                ".pw_ide": True,
                ".pw_ide.user.yaml": True,
                "bazel-bin": True,
                "bazel-out": True,
                "bazel-pigweed": True,
                "bazel-testlogs": True,
                "environment": True,
                "node_modules": True,
                "out": True,
            }
        ),
        "files.insertFinalNewline": True,
        "files.trimTrailingWhitespace": True,
        "search.useGlobalIgnoreFiles": True,
        "grunt.autoDetect": "off",
        "gulp.autoDetect": "off",
        "jake.autoDetect": "off",
        "npm.autoDetect": "off",
        "clangd.onConfigChanged": "restart",
        "C_Cpp.intelliSenseEngine": "Disabled",
        "[cpp]": OrderedDict(
            {"editor.defaultFormatter": "llvm-vs-code-extensions.vscode-clangd"}
        ),
        "python.analysis.diagnosticSeverityOverrides": OrderedDict(
            # Due to our project structure, the linter spuriously thinks we're
            # shadowing system modules any time we import them. This disables
            # that check.
            {"reportShadowedImports": "none"}
        ),
        # The "strict" mode is much more strict than what we currently enforce.
        "python.analysis.typeCheckingMode": "basic",
        "python.formatting.provider": "yapf",
        "python.linting.pylintEnabled": True,
        "python.linting.mypyEnabled": True,
        "python.testing.unittestEnabled": True,
        "[python]": OrderedDict({"editor.tabSize": 4}),
        "typescript.tsc.autoDetect": "off",
        "[gn]": OrderedDict({"editor.defaultFormatter": "msedge-dev.gnls"}),
        "[proto3]": OrderedDict(
            {"editor.defaultFormatter": "zxh404.vscode-proto3"}
        ),
    }
)

_DEFAULT_TASKS: EditorSettingsDict = OrderedDict(
    {
        "version": "2.0.0",
        "tasks": [
            {
                "type": "process",
                "label": "Pigweed: Format",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw format --fix'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: Presubmit",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw presubmit'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Set Python Virtual Environment",
                "command": "${command:python.setInterpreter}",
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Restart Python Language Server",
                "command": "${command:python.analysis.restartLanguageServer}",
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Restart C++ Language Server",
                "command": "${command:clangd.restart}",
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: Sync IDE",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide sync'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: Current C++ Target Toolchain",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide cpp'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": "Pigweed: List C++ Target Toolchains",
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide cpp --list'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "type": "process",
                "label": (
                    "Pigweed: Change C++ Target Toolchain "
                    "without LSP restart"
                ),
                "command": "${config:python.defaultInterpreterPath}",
                "args": [
                    "-m",
                    "pw_ide.activate",
                    "-x 'pw ide cpp --set ${input:availableTargetToolchains}'",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
            {
                "label": "Pigweed: Set C++ Target Toolchain",
                "dependsOrder": "sequence",
                "dependsOn": [
                    "Pigweed: Change C++ Target Toolchain without LSP restart",
                    "Pigweed: Restart C++ Language Server",
                ],
                "presentation": {
                    "focus": True,
                },
                "problemMatcher": [],
            },
        ],
    }
)

_DEFAULT_EXTENSIONS: EditorSettingsDict = OrderedDict(
    {
        "recommendations": [
            "llvm-vs-code-extensions.vscode-clangd",
            "ms-python.python",
            "npclaudiu.vscode-gn",
            "msedge-dev.gnls",
            "zxh404.vscode-proto3",
            "josetr.cmake-language-support-vscode",
            "swyddfa.esbonio",
        ],
        "unwantedRecommendations": [
            "ms-vscode.cpptools",
            "persidskiy.vscode-gnformat",
            "lextudio.restructuredtext",
            "trond-snekvik.simple-rst",
        ],
    }
)

_DEFAULT_LAUNCH: EditorSettingsDict = OrderedDict(
    {
        "version": "0.2.0",
        "configurations": [],
    }
)


def _default_settings(
    pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return OrderedDict(
        {
            **_DEFAULT_SETTINGS,
            **_local_terminal_integrated_env(),
            **_local_clangd_settings(pw_ide_settings),
            **_local_python_settings(),
        }
    )


def _default_tasks(
    pw_ide_settings: PigweedIdeSettings,
    state: Optional[CppIdeFeaturesState] = None,
) -> EditorSettingsDict:
    if state is None:
        state = CppIdeFeaturesState(pw_ide_settings)

    inputs = [
        {
            "type": "pickString",
            "id": "availableTargetToolchains",
            "description": "Available target toolchains",
            "options": list(state.targets),
        }
    ]

    return OrderedDict(**_DEFAULT_TASKS, inputs=inputs)


def _default_extensions(
    _pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return _DEFAULT_EXTENSIONS


def _default_launch(
    _pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return _DEFAULT_LAUNCH


DEFAULT_SETTINGS_PATH = Path(os.path.expandvars('$PW_PROJECT_ROOT')) / '.vscode'


class VscSettingsType(Enum):
    """Visual Studio Code settings files.

    VSC supports editor settings (``settings.json``), recommended
    extensions (``extensions.json``), tasks (``tasks.json``), and
    launch/debug configurations (``launch.json``).
    """

    SETTINGS = 'settings'
    TASKS = 'tasks'
    EXTENSIONS = 'extensions'
    LAUNCH = 'launch'

    @classmethod
    def all(cls) -> List['VscSettingsType']:
        return list(cls)


class VscSettingsManager(EditorSettingsManager[VscSettingsType]):
    """Manages all settings for Visual Studio Code."""

    default_settings_dir = DEFAULT_SETTINGS_PATH
    file_format = Json5FileFormat()

    types_with_defaults: EditorSettingsTypesWithDefaults = {
        VscSettingsType.SETTINGS: _default_settings,
        VscSettingsType.TASKS: _default_tasks,
        VscSettingsType.EXTENSIONS: _default_extensions,
        VscSettingsType.LAUNCH: _default_launch,
    }


def _prompt_for_path(reporter: StatusReporter) -> Path:
    reporter.info(
        [
            "Hmmm... I can't seem to find your Visual Studio Code binary path!",
            "You can provide it manually here, or Ctrl-C to cancel.",
        ]
    )

    path = Path(input("> "))

    if path.exists():
        return path

    reporter.err("Nothing found there!")
    raise FileNotFoundError()


# TODO(chadnorvell): Replace this when we support Python 3.11 with:
# _PathData = Tuple[Optional[str], *Tuple[str]]
_PathData = List[Optional[str]]


def _try_code_path(path_list: _PathData) -> Optional[Path]:
    root, *rest = path_list

    if root is None:
        return None

    path = Path(root)

    for part in rest:
        if part is None:
            return None

        path /= part

    return path


def _try_each_code_path(
    reporter: StatusReporter, *path_lists: _PathData
) -> Path:
    for path_list in path_lists:
        if (path := _try_code_path(path_list)) is not None:
            return path

    if (path_str := shutil.which('code')) is not None:
        return Path(path_str)

    return _prompt_for_path(reporter)


def _get_vscode_exe_path(
    reporter: StatusReporter, system: str = platform.system()
) -> Path:
    if system == 'Darwin':
        return _try_each_code_path(
            reporter,
            [
                '/Applications',
                'Visual Studio Code.app',
                'Contents',
                'Resources',
                'app',
                'bin',
                'code',
            ],
        )

    if system == 'Windows':
        return _try_each_code_path(
            reporter,
            [
                os.getenv('APPDATA'),
                'Local',
                'Programs',
                'Microsoft VS Code',
                'bin',
                'code.exe',
            ],
            [
                os.getenv('LOCALAPPDATA'),
                'Local',
                'Programs',
                'Microsoft VS Code',
                'bin',
                'code.exe',
            ],
        )

    if system == 'Linux':
        return _try_each_code_path(
            reporter,
            ['/usr', 'bin', 'code'],
            ['/usr', 'local', 'bin', 'code'],
        )

    return _prompt_for_path(reporter)


def _get_latest_extension_vsix() -> Path:
    return sorted(_VSIX_DIR.glob('*.vsix'), reverse=True)[0]


def install_extension_from_vsix(reporter: StatusReporter) -> None:
    """Install the latest pre-built VSC extension from its VSIX file.

    Normally, extensions are installed from the VSC extension marketplace.
    This will instead install the extension directly from a file.
    """
    extension_path = _get_latest_extension_vsix()
    vscode_exe_path = _get_vscode_exe_path(reporter)

    reporter.info(f"Path: {vscode_exe_path}")
    subprocess.run(
        [vscode_exe_path, '--install-extension', extension_path], check=True
    )



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Defines arguments for the pw command."""

import argparse
from dataclasses import dataclass, field
from enum import Enum
import logging
from pathlib import Path
import sys
from typing import List, NoReturn, Optional

from pw_cli import argument_types, plugins
from pw_cli.branding import banner

_HELP_HEADER = '''The Pigweed command line interface (CLI).

Example uses:
    pw logdemo
    pw --loglevel debug watch -C out
'''


def parse_args() -> argparse.Namespace:
    return arg_parser().parse_args()


class ShellCompletionFormat(Enum):
    """Supported shell tab completion modes."""

    BASH = 'bash'
    ZSH = 'zsh'


@dataclass(frozen=True)
class ShellCompletion:
    option_strings: List[str] = field(default_factory=list)
    help: Optional[str] = None
    choices: Optional[List[str]] = None
    flag: bool = True

    def bash_completion(self, text: str) -> List[str]:
        result: List[str] = []
        for option_str in self.option_strings:
            if option_str.startswith(text):
                result.append(option_str)
        return result

    def zsh_completion(self, text: str) -> List[str]:
        result: List[str] = []
        for option_str in self.option_strings:
            if option_str.startswith(text):
                short_and_long_opts = ' '.join(self.option_strings)
                # '(-h --help)-h[Display help message and exit]'
                # '(-h --help)--help[Display help message and exit]'
                help_text = self.help if self.help else ''
                state_str = ''
                if not self.flag:
                    state_str = ': :->' + option_str

                result.append(
                    f'({short_and_long_opts}){option_str}[{help_text}]'
                    f'{state_str}'
                )
        return result


def get_options_and_help(
    parser: argparse.ArgumentParser,
) -> List[ShellCompletion]:
    return list(
        ShellCompletion(
            option_strings=list(action.option_strings),
            help=action.help,
            choices=list(action.choices) if action.choices else [],
            flag=action.nargs == 0,
        )
        for action in parser._actions  # pylint: disable=protected-access
    )


def print_completions_for_option(
    parser: argparse.ArgumentParser,
    text: str = '',
    tab_completion_format: str = ShellCompletionFormat.BASH.value,
) -> None:
    matched_lines: List[str] = []
    for completion in get_options_and_help(parser):
        if tab_completion_format == ShellCompletionFormat.ZSH.value:
            matched_lines.extend(completion.zsh_completion(text))
        else:
            matched_lines.extend(completion.bash_completion(text))

    for line in matched_lines:
        print(line)


def print_banner() -> None:
    """Prints the PIGWEED (or project specific) banner to stderr."""
    print(banner() + '\n', file=sys.stderr)


def format_help(registry: plugins.Registry) -> str:
    """Returns the pw help information as a string."""
    return f'{arg_parser().format_help()}\n{registry.short_help()}'


class _ArgumentParserWithBanner(argparse.ArgumentParser):
    """Parser that the Pigweed banner when there are parsing errors."""

    def error(self, message: str) -> NoReturn:
        print_banner()
        self.print_usage(sys.stderr)
        self.exit(2, f'{self.prog}: error: {message}\n')


def add_tab_complete_arguments(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        '--tab-complete-option',
        nargs='?',
        help='Print tab completions for the supplied option text.',
    )
    parser.add_argument(
        '--tab-complete-format',
        choices=list(shell.value for shell in ShellCompletionFormat),
        default='bash',
        help='Output format for tab completion results.',
    )
    return parser


def arg_parser() -> argparse.ArgumentParser:
    """Creates an argument parser for the pw command."""
    argparser = _ArgumentParserWithBanner(
        prog='pw',
        add_help=False,
        description=_HELP_HEADER,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Do not use the built-in help argument so that displaying the help info can
    # be deferred until the pw plugins have been registered.
    argparser.add_argument(
        '-h',
        '--help',
        action='store_true',
        help='Display this help message and exit',
    )
    argparser.add_argument(
        '-C',
        '--directory',
        type=argument_types.directory,
        default=Path.cwd(),
        help='Change to this directory before doing anything',
    )
    argparser.add_argument(
        '-l',
        '--loglevel',
        type=argument_types.log_level,
        default=logging.INFO,
        help='Set the log level (debug, info, warning, error, critical)',
    )
    argparser.add_argument(
        '--debug-log',
        help=(
            'Additionally log to this file at debug level; does not affect '
            'terminal output'
        ),
    )
    argparser.add_argument(
        '--no-banner',
        action='store_true',
        help='Do not print the Pigweed banner',
    )
    argparser.add_argument(
        '--tab-complete-command',
        nargs='?',
        help='Print tab completions for the supplied command text.',
    )
    argparser.add_argument(
        'command',
        nargs='?',
        help='Which command to run; see supported commands below',
    )
    argparser.add_argument(
        'plugin_args',
        metavar='...',
        nargs=argparse.REMAINDER,
        help='Remaining arguments are forwarded to the command',
    )
    return add_tab_complete_arguments(argparser)



#!/usr/bin/env python3

# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Restructured Text Formatting."""

import argparse
from dataclasses import dataclass, field
import difflib
from functools import cached_property
from pathlib import Path
import textwrap
from typing import List, Optional

from pw_presubmit.tools import colorize_diff

DEFAULT_TAB_WIDTH = 8
CODE_BLOCK_INDENTATION = 3


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        '--tab-width',
        default=DEFAULT_TAB_WIDTH,
        help='Number of spaces to use when converting tab characters.',
    )
    parser.add_argument(
        '--diff',
        action='store_true',
        help='Print a diff of formatting changes.',
    )
    parser.add_argument(
        '-i',
        '--in-place',
        action='store_true',
        help='Replace existing file with the reformatted copy.',
    )
    parser.add_argument(
        'rst_files',
        nargs='+',
        default=[],
        type=Path,
        help='Paths to rst files.',
    )

    return parser.parse_args()


def _indent_amount(line: str) -> int:
    return len(line) - len(line.lstrip())


def _reindent(input_text: str, amount: int) -> str:
    text = ''
    for line in textwrap.dedent(input_text).splitlines():
        if len(line.strip()) == 0:
            text += '\n'
            continue
        text += ' ' * amount
        text += line
        text += '\n'
    return text


def _strip_trailing_whitespace(line: str) -> str:
    return line.rstrip() + '\n'


@dataclass
class CodeBlock:
    """Store a single code block."""

    directive_lineno: int
    directive_line: str
    first_line_indent: Optional[int] = None
    end_lineno: Optional[int] = None
    option_lines: List[str] = field(default_factory=list)
    code_lines: List[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        self._blank_line_after_options_found = False

    def finished(self) -> bool:
        return self.end_lineno is not None

    def append_line(self, index: int, line: str) -> None:
        """Process a line for this code block."""
        # Check if outside the code block (indentation is less):
        if (
            self.first_line_indent is not None
            and len(line.strip()) > 0
            and _indent_amount(line) < self.first_line_indent
        ):
            # Code block ended
            self.end_lineno = index
            return

        # If first line indent hasn't been found
        if self.first_line_indent is None:
            # Check if the first word is a directive option.
            # E.g. :caption:
            line_words = line.split()
            if (
                line_words
                and line_words[0].startswith(':')
                and line_words[0].endswith(':')
            ):
                self.option_lines.append(line.rstrip())
                return
            # Check for a blank line
            if len(line.strip()) == 0:
                if (
                    self.option_lines
                    and not self._blank_line_after_options_found
                ):
                    self._blank_line_after_options_found = True
                return
            # Check for a line that is a continuation of a previous option.
            if self.option_lines and not self._blank_line_after_options_found:
                self.option_lines.append(line.rstrip())
                return

            self.first_line_indent = _indent_amount(line)

        # Save this line as code.
        self.code_lines.append(line.rstrip())

    @cached_property
    def directive_indent_amount(self) -> int:
        return _indent_amount(self.directive_line)

    def options_block_text(self) -> str:
        return _reindent(
            input_text='\n'.join(self.option_lines),
            amount=self.directive_indent_amount + CODE_BLOCK_INDENTATION,
        )

    def code_block_text(self) -> str:
        return _reindent(
            input_text='\n'.join(self.code_lines),
            amount=self.directive_indent_amount + CODE_BLOCK_INDENTATION,
        )

    def to_text(self) -> str:
        text = ''
        text += self.directive_line
        if self.option_lines:
            text += self.options_block_text()
        text += '\n'
        text += self.code_block_text()
        text += '\n'
        return text

    def __repr__(self) -> str:
        return self.to_text()


def reindent_code_blocks(in_text: str) -> str:
    """Reindent code blocks to 3 spaces."""
    out_text = ''
    current_block: Optional[CodeBlock] = None
    for index, line in enumerate(in_text.splitlines(keepends=True)):
        # If a code block is active, process this line.
        if current_block:
            current_block.append_line(index, line)
            if current_block.finished():
                out_text += current_block.to_text()
                # This line wasn't part of the code block, process as normal.
                out_text += _strip_trailing_whitespace(line)
                # Erase this code_block variable
                current_block = None
        # Check for new code block start
        elif line.lstrip().startswith('.. code') and line.rstrip().endswith(
            '::'
        ):
            current_block = CodeBlock(
                directive_lineno=index, directive_line=line
            )
            continue
        else:
            out_text += _strip_trailing_whitespace(line)
    # If the document ends with a code block it may still need to be written.
    if current_block is not None:
        out_text += current_block.to_text()
    return out_text


def reformat_rst(
    file_name: Path,
    diff: bool = False,
    in_place: bool = False,
    tab_width: int = DEFAULT_TAB_WIDTH,
) -> List[str]:
    """Reformat an rst file.

    Returns a list of diff lines."""
    in_text = file_name.read_text()

    # Replace tabs with spaces
    out_text = in_text.replace('\t', ' ' * tab_width)

    # Indent .. code-block:: directives.
    out_text = reindent_code_blocks(in_text)

    result_diff = list(
        difflib.unified_diff(
            in_text.splitlines(True),
            out_text.splitlines(True),
            f'{file_name}  (original)',
            f'{file_name}  (reformatted)',
        )
    )
    if diff and result_diff:
        print(''.join(colorize_diff(result_diff)))

    if in_place:
        file_name.write_text(out_text)

    return result_diff


def rst_format_main(
    rst_files: List[Path],
    diff: bool = False,
    in_place: bool = False,
    tab_width: int = DEFAULT_TAB_WIDTH,
) -> None:
    for rst_file in rst_files:
        reformat_rst(rst_file, diff, in_place, tab_width)


if __name__ == '__main__':
    rst_format_main(**vars(_parse_args()))



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Presubmit to npm install and run tests"""

from pw_presubmit.presubmit import call
from pw_presubmit.presubmit_context import PresubmitContext


def npm_test(ctx: PresubmitContext) -> None:
    """Run npm install and npm test in Pigweed root to test all web modules"""
    call('npm', "install", cwd=ctx.root)
    call('npm', "test", cwd=ctx.root)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Default pw build script for upstream Pigweed."""

import logging
from pathlib import Path
import shutil
import sys

import pw_presubmit.pigweed_presubmit
from pw_presubmit.build import gn_args

from pw_build.build_recipe import (
    BuildCommand,
    BuildRecipe,
)
from pw_build.project_builder_presubmit_runner import (
    should_gn_gen,
    main,
)


_LOG = logging.getLogger('pw_build')


def pigweed_upstream_main() -> int:
    """Entry point for Pigweed upstream ``pw build`` command.

    Defines one or more BuildRecipes and passes that along with all of Pigweed
    upstream presubmit programs to the project_builder_presubmit_runner.main to
    start a pw build invocation.

    Returns:
      An int representing the success or failure status of the build; 0 if
      successful, 1 if failed.

    Command line usage examples:

    .. code-block:: bash
       pw build --list
       pw build --step gn_combined_build_check --step gn_python_*
       pw build --recipe default_ninja
       pw build --recipe default_* --watch
       pw build -C out --watch
    """
    default_bazel_targets = ['//...:all']
    bazel_root = Path('out/bazel')
    bazel_root.mkdir(parents=True, exist_ok=True)

    default_gn_gen_command = [
        'gn',
        'gen',
        '{build_dir}',
        '--export-compile-commands',
    ]
    if shutil.which('ccache'):
        default_gn_gen_command.append(gn_args(pw_command_launcher='ccache'))

    build_recipes = [
        BuildRecipe(
            build_dir=Path('out/gn'),
            title='default_ninja',
            steps=[
                BuildCommand(
                    run_if=should_gn_gen,
                    command=default_gn_gen_command,
                ),
                BuildCommand(
                    build_system_command='ninja',
                    targets=['default'],
                ),
            ],
        ),
        BuildRecipe(
            build_dir=bazel_root,
            title='default_bazel',
            steps=[
                BuildCommand(
                    build_system_command='bazel',
                    build_system_extra_args=[
                        'build',
                        '--verbose_failures',
                        '--worker_verbose',
                    ],
                    targets=default_bazel_targets,
                ),
                BuildCommand(
                    build_system_command='bazel',
                    build_system_extra_args=[
                        'test',
                        '--test_output=errors',
                    ],
                    targets=default_bazel_targets,
                ),
            ],
        ),
    ]

    return main(
        presubmit_programs=pw_presubmit.pigweed_presubmit.PROGRAMS,
        build_recipes=build_recipes,
        default_root_logfile=Path('out/build.txt'),
    )


if __name__ == '__main__':
    sys.exit(pigweed_upstream_main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_compiler_protos/test.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_protobuf_compiler_protos/test.proto',
  package='pw.protobuf_compiler.test',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&pw_protobuf_compiler_protos/test.proto\x12\x19pw.protobuf_compiler.test*\x18\n\x04\x45num\x12\x07\n\x03\x46OO\x10\x00\x12\x07\n\x03\x42\x41R\x10\x01\x62\x06proto3'
)

_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.protobuf_compiler.test.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FOO', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BAR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=69,
  serialized_end=93,
)
_sym_db.RegisterEnumDescriptor(_ENUM)

Enum = enum_type_wrapper.EnumTypeWrapper(_ENUM)
FOO = 0
BAR = 1


DESCRIPTOR.enum_types_by_name['Enum'] = _ENUM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Console Reusable UI widgets."""

# pylint: disable=unused-import
from pw_console.widgets.border import create_border
from pw_console.widgets.checkbox import (
    ToolbarButton,
    to_checkbox,
    to_setting,
    to_checkbox_with_keybind_indicator,
    to_keybind_indicator,
    to_checkbox_text,
)
from pw_console.widgets.mouse_handlers import on_click
from pw_console.widgets.window_pane import (
    FloatingWindowPane,
    WindowPane,
    WindowPaneHSplit,
)
from pw_console.widgets.window_pane_toolbar import WindowPaneToolbar



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Parses Bazel rules from a local Bazel workspace."""

import json
import logging
import os
import subprocess

from io import StringIO
from pathlib import Path, PurePosixPath
from typing import (
    Any,
    Callable,
    Dict,
    IO,
    Iterable,
    Iterator,
    List,
    Optional,
    Union,
)
from xml.etree import ElementTree

_LOG = logging.getLogger(__name__)


BazelValue = Union[bool, int, str, List[str], Dict[str, str]]


class ParseError(Exception):
    """Raised when a Bazel query returns data that can't be parsed."""


def parse_invalid(attr: Dict[str, Any]) -> BazelValue:
    """Raises an error that a type is unrecognized."""
    attr_type = attr['type']
    raise ParseError(f'unknown type: {attr_type}, expected one of {BazelValue}')


class BazelRule:
    """Represents a Bazel rule as parsed from the query results."""

    def __init__(self, label: str, kind: str) -> None:
        """Create a Bazel rule.

        Args:
            label: An absolute Bazel label corresponding to this rule.
            kind: The type of Bazel rule, e.g. cc_library.
        """
        if not label.startswith('//'):
            raise ParseError(f'invalid label: {label}')
        if ':' in label:
            parts = label.split(':')
            if len(parts) != 2:
                raise ParseError(f'invalid label: {label}')
            self._package = parts[0][2:]
            self._target = parts[1]
        else:
            self._package = str(label)[2:]
            self._target = PurePosixPath(label).name
        self._kind = kind

        self._attrs: Dict[str, BazelValue] = {}

    def package(self) -> str:
        """Returns the package portion of this rule's name."""
        return self._package

    def label(self) -> str:
        """Returns this rule's full target name."""
        return f'//{self._package}:{self._target}'

    def kind(self) -> str:
        """Returns this rule's target type."""
        return self._kind

    def parse(self, attrs: Iterable[Dict[str, Any]]) -> None:
        """Maps JSON data from a bazel query into this object.

        Args:
            attrs: A dictionary of attribute names and values for the Bazel
                rule. These should match the output of
                `bazel cquery ... --output=jsonproto`.
        """
        attr_parsers: Dict[str, Callable[[Dict[str, Any]], BazelValue]] = {
            'boolean': lambda attr: attr.get('booleanValue', False),
            'integer': lambda attr: int(attr.get('intValue', '0')),
            'string': lambda attr: attr.get('stringValue', ''),
            'string_list': lambda attr: attr.get('stringListValue', []),
            'label_list': lambda attr: attr.get('stringListValue', []),
            'string_dict': lambda attr: {
                p['key']: p['value'] for p in attr.get('stringDictValue', [])
            },
        }
        for attr in attrs:
            if 'explicitlySpecified' not in attr:
                continue
            if not attr['explicitlySpecified']:
                continue
            try:
                attr_name = attr['name']
            except KeyError:
                raise ParseError(
                    f'missing "name" in {json.dumps(attr, indent=2)}'
                )
            try:
                attr_type = attr['type'].lower()
            except KeyError:
                raise ParseError(
                    f'missing "type" in {json.dumps(attr, indent=2)}'
                )

            attr_parser = attr_parsers.get(attr_type, parse_invalid)
            self._attrs[attr_name] = attr_parser(attr)

    def has_attr(self, attr_name: str) -> bool:
        """Returns whether the rule has an attribute of the given name.

        Args:
            attr_name: The name of the attribute.
        """
        return attr_name in self._attrs

    def get_bool(self, attr_name: str) -> bool:
        """Gets the value of a boolean attribute.

        Args:
            attr_name: The name of the boolean attribute.
        """
        val = self._attrs.get(attr_name, False)
        assert isinstance(val, bool)
        return val

    def get_int(self, attr_name: str) -> int:
        """Gets the value of an integer attribute.

        Args:
            attr_name: The name of the integer attribute.
        """
        val = self._attrs.get(attr_name, 0)
        assert isinstance(val, int)
        return val

    def get_str(self, attr_name: str) -> str:
        """Gets the value of a string attribute.

        Args:
            attr_name: The name of the string attribute.
        """
        val = self._attrs.get(attr_name, '')
        assert isinstance(val, str)
        return val

    def get_list(self, attr_name: str) -> List[str]:
        """Gets the value of a string list attribute.

        Args:
            attr_name: The name of the string list attribute.
        """
        val = self._attrs.get(attr_name, [])
        assert isinstance(val, list)
        return val

    def get_dict(self, attr_name: str) -> Dict[str, str]:
        """Gets the value of a string list attribute.

        Args:
            attr_name: The name of the string list attribute.
        """
        val = self._attrs.get(attr_name, {})
        assert isinstance(val, dict)
        return val

    def set_attr(self, attr_name: str, value: BazelValue) -> None:
        """Sets the value of an attribute.

        Args:
            attr_name: The name of the attribute.
            value: The value to set.
        """
        self._attrs[attr_name] = value


class BazelWorkspace:
    """Represents a local instance of a Bazel repository.

    Attributes:
        root: Path to the local instance of a Bazel workspace.
        packages: Bazel packages mapped to their default visibility.
        repo: The name of Bazel workspace.
    """

    def __init__(self, pathname: Path) -> None:
        """Creates an object representing a Bazel workspace at the given path.

        Args:
            pathname: Path to the local instance of a Bazel workspace.
        """
        self.root: Path = pathname
        self.packages: Dict[str, str] = {}
        self.repo: str = os.path.basename(pathname)

    def get_rules(self, kind: str) -> Iterator[BazelRule]:
        """Returns rules matching the given kind, e.g. 'cc_library'."""
        self._load_packages()
        results = self._query(
            'cquery', f'kind({kind}, //...)', '--output=jsonproto'
        )
        json_data = json.loads(results)
        for result in json_data.get('results', []):
            rule_data = result['target']['rule']
            target = rule_data['name']
            rule = BazelRule(target, kind)
            default_visibility = self.packages[rule.package()]
            rule.set_attr('visibility', [default_visibility])
            rule.parse(rule_data['attribute'])
            yield rule

    def _load_packages(self) -> None:
        """Scans the workspace for packages and their default visibilities."""
        if self.packages:
            return
        packages = self._query('query', '//...', '--output=package').split('\n')
        for package in packages:
            results = self._query(
                'query', f'buildfiles(//{package}:*)', '--output=xml'
            )
            xml_data = ElementTree.fromstring(results)
            for pkg_elem in xml_data:
                if not pkg_elem.attrib['name'].startswith(f'//{package}:'):
                    continue
                for elem in pkg_elem:
                    if elem.tag == 'visibility-label':
                        self.packages[package] = elem.attrib['name']

    def _query(self, *args: str) -> str:
        """Invokes `bazel cquery` with the given selector."""
        output = StringIO()
        self._exec(*args, '--noshow_progress', output=output)
        return output.getvalue()

    def _exec(self, *args: str, output: Optional[IO] = None) -> None:
        """Execute a Bazel command in the workspace."""
        cmdline = ['bazel'] + list(args) + ['--noshow_progress']
        result = subprocess.run(
            cmdline,
            cwd=self.root,
            capture_output=True,
        )
        if not result.stdout:
            _LOG.error(result.stderr.decode('utf-8'))
            raise ParseError(f'Failed to query Bazel workspace: {self.root}')
        if output:
            output.write(result.stdout.decode('utf-8').strip())

    def run(self, label: str, *args, output: Optional[IO] = None) -> None:
        """Invokes `bazel run` on the given label.

        Args:
            label: A Bazel target, e.g. "@repo//package:target".
            args: Additional options to pass to `bazel`.
            output: Optional destination for the output of the command.
        """
        self._exec('run', label, *args, output=output)

    def revision(self) -> str:
        try:
            result = subprocess.run(
                ['git', 'rev-parse', 'HEAD'],
                cwd=self.root,
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as error:
            print(error.stderr.decode('utf-8'))
            raise
        return result.stdout.decode('utf-8').strip()

    def url(self) -> str:
        try:
            result = subprocess.run(
                ['git', 'remote', 'get-url', 'origin'],
                cwd=self.root,
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as error:
            print(error.stderr.decode('utf-8'))
            raise
        return result.stdout.decode('utf-8').strip()



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for manipulating GN labels and paths."""

from __future__ import annotations

import re

from pathlib import PurePosixPath
from typing import Optional, Union


class MalformedGnError(Exception):
    """Raised when creating a GN object fails."""


class GnPath:
    """Represents a GN source path to a file in the source tree."""

    def __init__(
        self,
        base: Union[str, PurePosixPath, GnPath],
        bazel: Optional[str] = None,
        gn: Optional[str] = None,  # pylint: disable=invalid-name
    ) -> None:
        """Creates a GN source path.

        Args:
            base: A base GN source path. Other parameters are used to define
                this object relative to the base.
            bazel: A Bazel path relative to `base`.
            gn: A GN source path relative to `base`.
        """
        self._path: PurePosixPath
        base_path = _as_path(base)
        if bazel:
            self._from_bazel(base_path, bazel)
        elif gn:
            self._from_gn(base_path, gn)
        else:
            self._path = base_path

    def __str__(self) -> str:
        return str(self._path)

    def _from_bazel(self, base_path: PurePosixPath, label: str) -> None:
        """Populates this object using a Bazel file label.

        A Bazel label looks like:
            //[{package-path}][:{relative-path}]
        e.g.
            "//foo"             => "$base/foo"
            "//:bar/baz.txt"    => "$base/bar/baz.txt"
            "//foo:bar/baz.txt" => "$base/foo/bar/baz.txt"
        """
        match = re.match(r'//([^():]*)(?::([^():]+))?', label)
        if not match:
            raise MalformedGnError(f'invalid path: {label}')
        groups = filter(None, match.groups())
        self._path = base_path.joinpath(*groups)

    def _from_gn(self, base_path: PurePosixPath, path: str) -> None:
        """Populates this object using a GN label.

        Source-relative paths interpreted relative to `base`. Source-absolute
        paths are used directly.
        """
        if path.startswith('//'):
            self._path = PurePosixPath(path)
        else:
            self._path = base_path.joinpath(path)

    def path(self) -> str:
        """Returns the object's path."""
        return str(self._path)

    def file(self) -> str:
        """Like GN's `get_path_info(..., "file")`."""
        return self._path.name

    def name(self) -> str:
        """Like GN's `get_path_info(..., "name")`."""
        return self._path.stem

    def extension(self) -> str:
        """Like GN's `get_path_info(..., "extension")`."""
        suffix = self._path.suffix
        return suffix[1:] if suffix.startswith('.') else suffix

    def dir(self) -> str:
        """Like GN's `get_path_info(..., "dir")`."""
        return str(self._path.parent)


class GnLabel:
    """Represents a GN dependency.

    See https://gn.googlesource.com/gn/+/main/docs/reference.md#labels.
    """

    def __init__(
        self,
        base: Union[str, PurePosixPath, GnLabel],
        public: bool = False,
        bazel: Optional[str] = None,
        gn: Optional[str] = None,  # pylint: disable=invalid-name
    ) -> None:
        """Creates a GN label.

        Args:
            base: A base GN label. Other parameters are used to define this
                object relative to the base.
            public: When this label is used to refer to a GN `dep`, this flag
                indicates if it should be a `public_dep`.
            bazel: A Bazel label relative to `base`.
            gn: A GN label relative to `base`.
        """
        self._name: str
        self._path: PurePosixPath
        self._toolchain: Optional[str] = None
        self._public: bool = public
        self._repo: Optional[str] = None
        base_path = _as_path(base)
        if bazel:
            self._from_bazel(base_path, bazel)
        elif gn:
            self._from_gn(base_path, gn)
        elif ':' in str(base_path):
            parts = str(base_path).split(':')
            self._path = PurePosixPath(':'.join(parts[:-1]))
            self._name = parts[-1]
        elif isinstance(base, GnLabel):
            self._path = base._path
            self._name = base._name
        else:
            self._path = base_path
            self._name = self._path.name

    def __str__(self):
        return self.with_toolchain() if self._toolchain else self.no_toolchain()

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def _from_bazel(self, base: PurePosixPath, label: str):
        """Populates this object using a Bazel label."""
        match = re.match(r'(?:@([^():/]*))?//(.+)', label)
        if not match:
            raise MalformedGnError(f'invalid label: {label}')
        self._repo = match[1]
        if self._repo:
            self._from_gn(PurePosixPath('$repo'), match[2])
        else:
            self._from_gn(base, match[2])

    def _from_gn(self, base: PurePosixPath, label: str):
        """Populates this object using a GN label."""
        if label.startswith('//') or label.startswith('$'):
            path = label
        else:
            path = str(base.joinpath(label))
        if ':' in path:
            parts = path.split(':')
            self._path = PurePosixPath(':'.join(parts[:-1]))
            self._name = parts[-1]
        else:
            self._path = PurePosixPath(path)
            self._name = self._path.name
        parts = []
        for part in self._path.parts:
            if part == '..' and parts and parts[-1] != '..':
                parts.pop()
            else:
                parts.append(part)
        self._path = PurePosixPath(*parts)

    def name(self) -> str:
        """Like GN's `get_label_info(..., "name"`)."""
        return self._name

    def dir(self) -> str:
        """Like GN's `get_label_info(..., "dir"`)."""
        return str(self._path)

    def no_toolchain(self) -> str:
        """Like GN's `get_label_info(..., "label_no_toolchain"`)."""
        if self._path == PurePosixPath():
            return f':{self._name}'
        name = f':{self._name}' if self._name != self._path.name else ''
        return f'{self._path}{name}'

    def with_toolchain(self) -> str:
        """Like GN's `get_label_info(..., "label_with_toolchain"`)."""
        toolchain = self._toolchain if self._toolchain else 'default_toolchain'
        if self._path == PurePosixPath():
            return f':{self._name}({toolchain})'
        name = f':{self._name}' if self._name != self._path.name else ''
        return f'{self._path}{name}({toolchain})'

    def public(self) -> bool:
        """Returns whether this is a public dep."""
        return self._public

    def repo(self) -> str:
        """Returns the label's repo, if any."""
        return self._repo or ''

    def resolve_repo(self, repo: str) -> None:
        """Replaces the repo placeholder with the given value."""
        if self._path and self._path.parts[0] == '$repo':
            self._path = PurePosixPath(
                '$dir_pw_third_party', repo, *self._path.parts[1:]
            )

    def relative_to(self, start: Union[str, PurePosixPath, GnLabel]) -> str:
        """Returns a label string relative to the given starting label."""
        start_path = _as_path(start)
        if not start:
            return self.no_toolchain()
        if self._path == start_path:
            return f':{self._name}'
        path = _relative_to(self._path, start_path)
        name = f':{self._name}' if self._name != self._path.name else ''
        return f'{path}{name}'

    def joinlabel(self, relative: str) -> GnLabel:
        """Creates a new label by extending the current label."""
        return GnLabel(self._path.joinpath(relative))


class GnVisibility:
    """Represents a GN visibility scope."""

    def __init__(
        self,
        base: Union[str, PurePosixPath, GnLabel],
        label: Union[str, PurePosixPath, GnLabel],
        bazel: Optional[str] = None,
        gn: Optional[str] = None,  # pylint: disable=invalid-name
    ) -> None:
        """Creates a GN visibility scope.

        Args:
            base: A base GN label. Other parameters are used to define this
                object relative to the base.
            label: The label of the directory in which this scope is being
                defined.
            bazel: An absolute Bazel visibility label.
            gn: A GN visibility label.
        """
        self._scope: GnLabel
        label_path = _as_path(label)
        if bazel:
            self._from_bazel(_as_path(base), label_path, bazel)
        elif gn:
            self._from_gn(label_path, gn)
        else:
            self._scope = GnLabel(label)

    def __str__(self):
        return str(self._scope)

    def _from_bazel(
        self, base: PurePosixPath, label: PurePosixPath, scope: str
    ):
        """Populates this object using a Bazel visibility label."""
        if scope == '//visibility:public':
            self._scope = GnLabel('//*')
        elif scope == '//visibility:private':
            self._scope = GnLabel(label, gn=':*')
        elif not (match := re.match(r'//([^():]*):([^():]+)', scope)):
            raise MalformedGnError(f'invalid visibility scope: {scope}')
        elif match[2] == '__subpackages__':
            self._scope = GnLabel(base, gn=f'{match[1]}/*')
        elif match[2] == '__pkg__':
            self._scope = GnLabel(base, gn=f'{match[1]}:*')
        else:
            raise MalformedGnError(f'unsupported visibility scope: {scope}')

    def _from_gn(self, label: PurePosixPath, scope: str):
        """Populates this object using a GN visibility scope."""
        self._scope = GnLabel(label, gn=scope)

    def relative_to(self, start: Union[str, PurePosixPath, GnLabel]) -> str:
        """Returns a label string relative to the given starting label."""
        return self._scope.relative_to(start)

    def within(self, other: GnVisibility) -> bool:
        """Returns whether this scope is a subset of another."""
        as_label = GnLabel(str(other))
        if as_label.name() == '*':
            _path = self._scope.dir()
            other_path = as_label.dir()
            if other_path == '//*':
                return True
            if other_path.endswith('*'):
                parent = PurePosixPath(other_path).parent
                return _is_relative_to(_path, parent)
            return _path == other_path
        return str(self) == str(other)


def _as_path(item: Union[str, GnPath, GnLabel, PurePosixPath]) -> PurePosixPath:
    """Converts an argument to be a PurePosixPath.

    Args:
        label: A string, path, or label to be converted to a PurePosixPath.
    """
    if isinstance(item, str):
        return PurePosixPath(item)
    if isinstance(item, GnPath):
        return PurePosixPath(item.path())
    if isinstance(item, GnLabel):
        return PurePosixPath(item.dir())
    return item


def _is_relative_to(
    path: Union[str, PurePosixPath], start: Union[str, PurePosixPath]
) -> bool:
    """Returns whether `path` is a subdirectory of `start`."""
    _path = PurePosixPath(path)
    _start = PurePosixPath(start)
    try:
        _path.relative_to(_start)
        return True
    except ValueError:
        return False


def _relative_to(
    path: Union[str, PurePosixPath], start: Union[str, PurePosixPath]
) -> PurePosixPath:
    """Like `PosixPath._relative_to`, but can ascend directories as well."""
    if not start:
        return PurePosixPath(path)
    _path = PurePosixPath(path)
    _start = PurePosixPath(start)
    if _path.parts[0] != _start.parts[0]:
        return _path
    ascend = PurePosixPath()
    while not _is_relative_to(_path, _start):
        if _start.parent == PurePosixPath():
            break
        _start = _start.parent
        ascend = ascend.joinpath('..')
    return ascend.joinpath(_path.relative_to(_start))



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Wrapers for socket clients to log read and write data."""
from __future__ import annotations

from typing import Callable, Optional, TYPE_CHECKING, Tuple, Union

import errno
import socket

from pw_console.plugins.bandwidth_toolbar import SerialBandwidthTracker

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer


class SocketClient:
    """Socket transport implementation."""

    FILE_SOCKET_SERVER = 'file'
    DEFAULT_SOCKET_SERVER = 'localhost'
    DEFAULT_SOCKET_PORT = 33000
    PW_RPC_MAX_PACKET_SIZE = 256

    def __init__(
        self,
        config: str,
        on_disconnect: Optional[Callable[[SocketClient], None]] = None,
    ):
        self._connection_type: int
        self._interface: Union[str, Tuple[str, int]]
        if config == 'default':
            self._connection_type = socket.AF_INET
            self._interface = (
                self.DEFAULT_SOCKET_SERVER,
                self.DEFAULT_SOCKET_PORT,
            )
        else:
            socket_server, socket_port_or_file = config.split(':')
            if socket_server == self.FILE_SOCKET_SERVER:
                # Unix socket support is available on Windows 10 since April
                # 2018. However, there is no Python support on Windows yet.
                # See https://bugs.python.org/issue33408 for more information.
                if not hasattr(socket, 'AF_UNIX'):
                    raise TypeError(
                        'Unix sockets are not supported in this environment.'
                    )
                self._connection_type = (
                    socket.AF_UNIX  # pylint: disable=no-member
                )
                self._interface = socket_port_or_file
            else:
                self._connection_type = socket.AF_INET
                self._interface = (socket_server, int(socket_port_or_file))

        self._on_disconnect = on_disconnect
        self._connected = False
        self.connect()

    def __del__(self):
        self.socket.close()

    def write(self, data: ReadableBuffer) -> None:
        """Writes data and detects disconnects."""
        if not self._connected:
            raise Exception('Socket is not connected.')
        try:
            self.socket.sendall(data)
        except socket.error as exc:
            if isinstance(exc.args, tuple) and exc.args[0] == errno.EPIPE:
                self._handle_disconnect()
            else:
                raise exc

    def read(self, num_bytes: int = PW_RPC_MAX_PACKET_SIZE) -> bytes:
        """Blocks until data is ready and reads up to num_bytes."""
        if not self._connected:
            raise Exception('Socket is not connected.')
        data = self.socket.recv(num_bytes)
        # Since this is a blocking read, no data returned means the socket is
        # closed.
        if not data:
            self._handle_disconnect()
        return data

    def connect(self) -> None:
        """Connects to socket."""
        self.socket = socket.socket(self._connection_type, socket.SOCK_STREAM)

        # Enable reusing address and port for reconnections.
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(socket, 'SO_REUSEPORT'):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self.socket.connect(self._interface)
        self._connected = True

    def _handle_disconnect(self):
        """Escalates a socket disconnect to the user."""
        self.socket.close()
        self._connected = False
        if self._on_disconnect:
            self._on_disconnect(self)


class SocketClientWithLogging(SocketClient):
    """Socket with read and write wrappers for logging."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bandwidth_tracker = SerialBandwidthTracker()

    def read(
        self, num_bytes: int = SocketClient.PW_RPC_MAX_PACKET_SIZE
    ) -> bytes:
        data = super().read(num_bytes)
        self._bandwidth_tracker.track_read_data(data)
        return data

    def write(self, data: ReadableBuffer) -> None:
        self._bandwidth_tracker.track_write_data(data)
        super().write(data)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Device classes to interact with targets via RPC."""

import logging
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, List, Union, Optional

from pw_hdlc.rpc import HdlcRpcClient, channel_output
from pw_hdlc.rpc import NoEncodingSingleChannelRpcClient, RpcClient
from pw_log.log_decoder import (
    Log,
    LogStreamDecoder,
    log_decoded_log,
    timestamp_parser_ns_since_boot,
)
from pw_log_rpc.rpc_log_stream import LogStreamHandler
from pw_metric import metric_parser
from pw_rpc import callback_client, Channel, console_tools
from pw_thread.thread_analyzer import ThreadSnapshotAnalyzer
from pw_thread_protos import thread_pb2
from pw_tokenizer import detokenize
from pw_tokenizer.proto import decode_optionally_tokenized
from pw_unit_test.rpc import run_tests as pw_unit_test_run_tests, TestRecord

# Internal log for troubleshooting this tool (the console).
_LOG = logging.getLogger('tools')
DEFAULT_DEVICE_LOGGER = logging.getLogger('rpc_device')


class Device:
    """Represents an RPC Client for a device running a Pigweed target.

    The target must have and RPC support, RPC logging.
    Note: use this class as a base for specialized device representations.
    """

    def __init__(
        self,
        channel_id: int,
        read,
        write,
        proto_library: List[Union[ModuleType, Path]],
        detokenizer: Optional[detokenize.Detokenizer] = None,
        timestamp_decoder: Optional[Callable[[int], str]] = None,
        rpc_timeout_s: float = 5,
        use_rpc_logging: bool = True,
        use_hdlc_encoding: bool = True,
    ):
        self.channel_id = channel_id
        self.protos = proto_library
        self.detokenizer = detokenizer
        self.rpc_timeout_s = rpc_timeout_s

        self.logger = DEFAULT_DEVICE_LOGGER
        self.logger.setLevel(logging.DEBUG)  # Allow all device logs through.

        callback_client_impl = callback_client.Impl(
            default_unary_timeout_s=self.rpc_timeout_s,
            default_stream_timeout_s=None,
        )

        def detokenize_and_log_output(data: bytes, _detokenizer=None):
            log_messages = data.decode(
                encoding='utf-8', errors='surrogateescape'
            )

            if self.detokenizer:
                log_messages = decode_optionally_tokenized(
                    self.detokenizer, data
                )

            for line in log_messages.splitlines():
                self.logger.info(line)

        self.client: RpcClient
        if use_hdlc_encoding:
            channels = [Channel(self.channel_id, channel_output(write))]
            self.client = HdlcRpcClient(
                read,
                self.protos,
                channels,
                detokenize_and_log_output,
                client_impl=callback_client_impl,
            )
        else:
            channel = Channel(self.channel_id, write)
            self.client = NoEncodingSingleChannelRpcClient(
                read,
                self.protos,
                channel,
                client_impl=callback_client_impl,
            )

        if use_rpc_logging:
            # Create the log decoder used by the LogStreamHandler.

            def decoded_log_handler(log: Log) -> None:
                log_decoded_log(log, self.logger)

            self._log_decoder = LogStreamDecoder(
                decoded_log_handler=decoded_log_handler,
                detokenizer=self.detokenizer,
                source_name='RpcDevice',
                timestamp_parser=(
                    timestamp_decoder
                    if timestamp_decoder
                    else timestamp_parser_ns_since_boot
                ),
            )

            # Start listening to logs as soon as possible.
            self.log_stream_handler = LogStreamHandler(
                self.rpcs, self._log_decoder
            )
            self.log_stream_handler.listen_to_logs()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def close(self) -> None:
        self.client.close()

    def info(self) -> console_tools.ClientInfo:
        return console_tools.ClientInfo('device', self.rpcs, self.client.client)

    @property
    def rpcs(self) -> Any:
        """Returns an object for accessing services on the specified channel."""
        return next(iter(self.client.client.channels())).rpcs

    def run_tests(self, timeout_s: Optional[float] = 5) -> TestRecord:
        """Runs the unit tests on this device."""
        return pw_unit_test_run_tests(self.rpcs, timeout_s=timeout_s)

    def get_and_log_metrics(self) -> dict:
        """Retrieves the parsed metrics and logs them to the console."""
        metrics = metric_parser.parse_metrics(
            self.rpcs, self.detokenizer, self.rpc_timeout_s
        )

        def print_metrics(metrics, path):
            """Traverses dictionaries, until a non-dict value is reached."""
            for path_name, metric in metrics.items():
                if isinstance(metric, dict):
                    print_metrics(metric, path + '/' + path_name)
                else:
                    _LOG.info('%s/%s: %s', path, path_name, str(metric))

        print_metrics(metrics, '')
        return metrics

    def snapshot_peak_stack_usage(self, thread_name: Optional[str] = None):
        snapshot_service = self.rpcs.pw.thread.proto.ThreadSnapshotService
        _, rsp = snapshot_service.GetPeakStackUsage(name=thread_name)

        thread_info = thread_pb2.SnapshotThreadInfo()
        for thread_info_block in rsp:
            for thread in thread_info_block.threads:
                thread_info.threads.append(thread)
        for line in str(ThreadSnapshotAnalyzer(thread_info)).splitlines():
            _LOG.info('%s', line)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for using HDLC with pw_rpc."""

from concurrent.futures import ThreadPoolExecutor
import io
import logging
from queue import SimpleQueue
import sys
import threading
import time
import socket
import subprocess
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from pw_protobuf_compiler import python_protos
import pw_rpc
from pw_rpc import callback_client

from pw_hdlc.decode import Frame, FrameDecoder
from pw_hdlc import encode

_LOG = logging.getLogger('pw_hdlc.rpc')

STDOUT_ADDRESS = 1
DEFAULT_ADDRESS = ord('R')
DEFAULT_CHANNEL_ID = 1
_VERBOSE = logging.DEBUG - 1


def channel_output(
    writer: Callable[[bytes], Any],
    address: int = DEFAULT_ADDRESS,
    delay_s: float = 0,
) -> Callable[[bytes], None]:
    """Returns a function that can be used as a channel output for pw_rpc."""

    if delay_s:

        def slow_write(data: bytes) -> None:
            """Slows down writes in case unbuffered serial is in use."""
            for byte in data:
                time.sleep(delay_s)
                writer(bytes([byte]))

        return lambda data: slow_write(encode.ui_frame(address, data))

    def write_hdlc(data: bytes):
        frame = encode.ui_frame(address, data)
        _LOG.log(_VERBOSE, 'Write %2d B: %s', len(frame), frame)
        writer(frame)

    return write_hdlc


FrameHandlers = Dict[int, Callable[[Frame], Any]]
FrameTypeT = TypeVar('FrameTypeT')


class DataReaderAndExecutor:
    """Reads incoming bytes, data processor that delegates frame handling.

    Executing callbacks in a ThreadPoolExecutor decouples reading the input
    stream from handling the data. That way, if a handler function takes a
    long time or crashes, this reading thread is not interrupted.
    """

    def __init__(
        self,
        read: Callable[[], bytes],
        on_read_error: Callable[[Exception], None],
        data_processor: Callable[[bytes], Iterable[FrameTypeT]],
        frame_handler: Callable[[FrameTypeT], None],
        handler_threads: Optional[int] = 1,
    ):
        """Creates the data reader and frame delegator.

        Args:
            read: Reads incoming bytes from the given transport, blocking until
              data is available or an exception is raised. Otherwise the reader
              will exit.
            on_read_error: Called when there is an error reading incoming bytes.
            data_processor: Processes read bytes and returns a frame-like object
              that the frame_handler can process.
            frame_handler: Handles a received frame.
            handler_threads: The number of threads in the executor pool.
        """

        self._read = read
        self._on_read_error = on_read_error
        self._data_processor = data_processor
        self._frame_handler = frame_handler
        self._handler_threads = handler_threads

        self._reader_thread = threading.Thread(
            target=self._run,
            # TODO: b/294858483 - When we are confident that we can cancel the
            # blocking read(), this no longer needs to be a daemon thread.
            daemon=True,
        )
        self._reader_stop = threading.Event()

    def start(self) -> None:
        """Starts the reading process."""
        self._reader_stop.clear()
        self._reader_thread.start()

    def stop(self) -> None:
        """Requests that the reading process stop.

        The thread will not stop immediately, but only after the ongoing read()
        operation completes or raises an exception.
        """
        self._reader_stop.set()
        # TODO: b/294858483 - When we are confident that we can cancel the
        # blocking read(), wait for the thread to exit.
        # self._reader_thread.join()

    def _run(self) -> None:
        """Reads raw data in a background thread."""
        with ThreadPoolExecutor(max_workers=self._handler_threads) as executor:
            while not self._reader_stop.is_set():
                try:
                    data = self._read()
                except Exception as exc:  # pylint: disable=broad-except
                    # Don't report the read error if the thread is stopping.
                    # The stream or device backing _read was likely closed,
                    # so errors are expected.
                    if not self._reader_stop.is_set():
                        self._on_read_error(exc)
                    _LOG.debug(
                        'DataReaderAndExecutor thread exiting due to exception',
                        exc_info=exc,
                    )
                    return

                if not data:
                    continue

                _LOG.log(_VERBOSE, 'Read %2d B: %s', len(data), data)

                for frame in self._data_processor(data):
                    executor.submit(self._frame_handler, frame)


# Writes to stdout by default, but sys.stdout.buffer is not guaranteed to exist
# (see https://docs.python.org/3/library/io.html#io.TextIOBase.buffer). Defer
# to sys.__stdout__.buffer if sys.stdout is wrapped with something that does not
# offer it.
def write_to_file(
    data: bytes,
    output: BinaryIO = getattr(sys.stdout, 'buffer', sys.__stdout__.buffer),
) -> None:
    output.write(data + b'\n')
    output.flush()


def default_channels(write: Callable[[bytes], Any]) -> List[pw_rpc.Channel]:
    return [pw_rpc.Channel(DEFAULT_CHANNEL_ID, channel_output(write))]


PathsModulesOrProtoLibrary = Union[
    Iterable[python_protos.PathOrModule], python_protos.Library
]


class RpcClient:
    """An RPC client with configurable incoming data processing."""

    def __init__(
        self,
        reader: DataReaderAndExecutor,
        paths_or_modules: PathsModulesOrProtoLibrary,
        channels: Iterable[pw_rpc.Channel],
        client_impl: Optional[pw_rpc.client.ClientImpl] = None,
    ):
        """Creates an RPC client.

        Args:
          read: Function that reads bytes; e.g serial_device.read.
          paths_or_modules: paths to .proto files or proto modules.
          channels: RPC channels to use for output.
          client_impl: The RPC Client implementation. Defaults to the callback
            client implementation if not provided.
        """
        if isinstance(paths_or_modules, python_protos.Library):
            self.protos = paths_or_modules
        else:
            self.protos = python_protos.Library.from_paths(paths_or_modules)

        if client_impl is None:
            client_impl = callback_client.Impl()

        self.client = pw_rpc.Client.from_modules(
            client_impl, channels, self.protos.modules()
        )

        # Start background thread that reads and processes RPC packets.
        self._reader = reader
        self._reader.start()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def close(self) -> None:
        self._reader.stop()

    def rpcs(self, channel_id: Optional[int] = None) -> Any:
        """Returns object for accessing services on the specified channel.

        This skips some intermediate layers to make it simpler to invoke RPCs
        from an HdlcRpcClient. If only one channel is in use, the channel ID is
        not necessary.
        """
        if channel_id is None:
            return next(iter(self.client.channels())).rpcs

        return self.client.channel(channel_id).rpcs

    def handle_rpc_packet(self, packet: bytes) -> None:
        if not self.client.process_packet(packet):
            _LOG.error('Packet not handled by RPC client: %s', packet)


class HdlcRpcClient(RpcClient):
    """An RPC client configured to run over HDLC.

    Expects HDLC frames to have addresses that dictate how to parse the HDLC
    payloads.
    """

    def __init__(
        self,
        read: Callable[[], bytes],
        paths_or_modules: PathsModulesOrProtoLibrary,
        channels: Iterable[pw_rpc.Channel],
        output: Callable[[bytes], Any] = write_to_file,
        client_impl: Optional[pw_rpc.client.ClientImpl] = None,
        *,
        _incoming_packet_filter_for_testing: Optional[
            pw_rpc.ChannelManipulator
        ] = None,
        rpc_frames_address: int = DEFAULT_ADDRESS,
        log_frames_address: int = STDOUT_ADDRESS,
        extra_frame_handlers: Optional[FrameHandlers] = None,
    ):
        """Creates an RPC client configured to communicate using HDLC.

        Args:
          read: Function that reads bytes; e.g serial_device.read.
          paths_or_modules: paths to .proto files or proto modules.
          channels: RPC channels to use for output.
          output: where to write "stdout" output from the device.
          client_impl: The RPC Client implementation. Defaults to the callback
            client implementation if not provided.
          rpc_frames_address: the address used in the HDLC frames for RPC
            packets. This can be the channel ID, or any custom address.
          log_frames_address: the address used in the HDLC frames for "stdout"
            output from the device.
          extra_fram_handlers: Optional mapping of HDLC frame addresses to their
            callbacks.
        """
        # Set up frame handling.
        rpc_output: Callable[[bytes], Any] = self.handle_rpc_packet
        if _incoming_packet_filter_for_testing is not None:
            _incoming_packet_filter_for_testing.send_packet = rpc_output
            rpc_output = _incoming_packet_filter_for_testing

        frame_handlers: FrameHandlers = {
            rpc_frames_address: lambda frame: rpc_output(frame.data),
            log_frames_address: lambda frame: output(frame.data),
        }
        if extra_frame_handlers:
            frame_handlers.update(extra_frame_handlers)

        def handle_frame(frame: Frame) -> None:
            # Suppress raising any frame errors to avoid crashes on data
            # processing, which may hide or drop other data.
            try:
                if not frame.ok():
                    _LOG.error('Failed to parse frame: %s', frame.status.value)
                    _LOG.debug('%s', frame.data)
                    return

                try:
                    frame_handlers[frame.address](frame)
                except KeyError:
                    _LOG.warning(
                        'Unhandled frame for address %d: %s',
                        frame.address,
                        frame,
                    )
            except:  # pylint: disable=bare-except
                _LOG.exception('Exception in HDLC frame handler thread')

        decoder = FrameDecoder()

        def on_read_error(exc: Exception) -> None:
            _LOG.error('data reader encountered an error', exc_info=exc)

        reader = DataReaderAndExecutor(
            read, on_read_error, decoder.process_valid_frames, handle_frame
        )
        super().__init__(reader, paths_or_modules, channels, client_impl)


class NoEncodingSingleChannelRpcClient(RpcClient):
    """An RPC client without any frame encoding with a single channel output.

    The caveat is that the provided read function must read entire frames.
    """

    def __init__(
        self,
        read: Callable[[], bytes],
        paths_or_modules: PathsModulesOrProtoLibrary,
        channel: pw_rpc.Channel,
        client_impl: Optional[pw_rpc.client.ClientImpl] = None,
    ):
        """Creates an RPC client over a single channel with no frame encoding.

        Args:
          read: Function that reads bytes; e.g serial_device.read.
          paths_or_modules: paths to .proto files or proto modules.
          channel: RPC channel to use for output.
          client_impl: The RPC Client implementation. Defaults to the callback
            client implementation if not provided.
        """

        def process_data(data: bytes):
            yield data

        def on_read_error(exc: Exception) -> None:
            _LOG.error('data reader encountered an error', exc_info=exc)

        reader = DataReaderAndExecutor(
            read, on_read_error, process_data, self.handle_rpc_packet
        )
        super().__init__(reader, paths_or_modules, [channel], client_impl)


def _try_connect(port: int, attempts: int = 10) -> socket.socket:
    """Tries to connect to the specified port up to the given number of times.

    This is helpful when connecting to a process that was started by this
    script. The process may need time to start listening for connections, and
    that length of time can vary. This retries with a short delay rather than
    having to wait for the worst case delay every time.
    """
    timeout_s = 0.001
    while True:
        time.sleep(timeout_s)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(('localhost', port))
            return sock
        except ConnectionRefusedError:
            sock.close()
            attempts -= 1
            if attempts <= 0:
                raise

            timeout_s *= 2


class SocketSubprocess:
    """Executes a subprocess and connects to it with a socket."""

    def __init__(self, command: Sequence, port: int) -> None:
        self._server_process = subprocess.Popen(command, stdin=subprocess.PIPE)
        self.stdin = self._server_process.stdin

        try:
            self.socket: socket.socket = _try_connect(port)  # 🧦
        except:
            self._server_process.terminate()
            self._server_process.communicate()
            raise

    def close(self) -> None:
        try:
            self.socket.close()
        finally:
            self._server_process.terminate()
            self._server_process.communicate()

    def __enter__(self) -> 'SocketSubprocess':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()


class HdlcRpcLocalServerAndClient:
    """Runs an RPC server in a subprocess and connects to it over a socket.

    This can be used to run a local RPC server in an integration test.
    """

    def __init__(
        self,
        server_command: Sequence,
        port: int,
        protos: PathsModulesOrProtoLibrary,
        *,
        incoming_processor: Optional[pw_rpc.ChannelManipulator] = None,
        outgoing_processor: Optional[pw_rpc.ChannelManipulator] = None,
    ) -> None:
        """Creates a new HdlcRpcLocalServerAndClient."""

        self.server = SocketSubprocess(server_command, port)

        self._bytes_queue: 'SimpleQueue[bytes]' = SimpleQueue()
        self._read_thread = threading.Thread(target=self._read_from_socket)
        self._read_thread.start()

        self.output = io.BytesIO()

        self.channel_output: Any = self.server.socket.sendall

        self._incoming_processor = incoming_processor
        if outgoing_processor is not None:
            outgoing_processor.send_packet = self.channel_output
            self.channel_output = outgoing_processor

        self.client = HdlcRpcClient(
            self._bytes_queue.get,
            protos,
            default_channels(self.channel_output),
            self.output.write,
            _incoming_packet_filter_for_testing=incoming_processor,
        ).client

    def _read_from_socket(self):
        while True:
            data = self.server.socket.recv(4096)
            self._bytes_queue.put(data)
            if not data:
                return

    def close(self):
        self.server.close()
        self.output.close()
        self._read_thread.join()

    def __enter__(self) -> 'HdlcRpcLocalServerAndClient':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Writes a formatted BUILD.gn _file."""

import os
import subprocess

from datetime import datetime
from pathlib import Path, PurePath, PurePosixPath
from types import TracebackType
from typing import Dict, IO, Iterable, Iterator, List, Optional, Type, Union

from pw_build.gn_config import GnConfig, GN_CONFIG_FLAGS
from pw_build.gn_target import GnTarget
from pw_build.gn_utils import GnLabel, GnPath, MalformedGnError

COPYRIGHT_HEADER = f'''
# Copyright {datetime.now().year} The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# DO NOT MANUALLY EDIT!'''


class GnWriter:
    """Represents a partial BUILD.gn file being constructed.

    Except for testing , callers should prefer using `GnFile`. That
    object wraps this one, and ensures that the GN file produced includes the
    relevant copyright header and is formatted correctly.

    Attributes:
        repos: A mapping of repository names to build args. These are used to
            replace repository names when writing labels.
        aliases: A mapping of label names to build args. These can be used to
            rewrite labels with alternate names, e.g. "gtest" to "googletest".
    """

    def __init__(self, file: IO) -> None:
        self._file: IO = file
        self._scopes: List[str] = []
        self._margin: str = ''
        self._needs_blank: bool = False
        self.repos: Dict[str, str] = {}
        self.aliases: Dict[str, str] = {}

    def write_comment(self, comment: Optional[str] = None) -> None:
        """Adds a GN comment.

        Args:
            comment: The comment string to write.
        """
        if not comment:
            self.write('#')
            return
        while len(comment) > 78:
            index = comment.rfind(' ', 0, 78)
            if index < 0:
                break
            self.write(f'# {comment[:index]}')
            comment = comment[index + 1 :]
        self.write(f'# {comment}')

    def write_import(self, gni: Union[str, PurePosixPath, GnPath]) -> None:
        """Adds a GN import.

        Args:
            gni: The source-relative path to a GN import file.
        """
        self._needs_blank = False
        self.write(f'import("{str(gni)}")')
        self._needs_blank = True

    def write_imports(self, imports: Iterable[str]) -> None:
        """Adds a list of GN imports.

        Args:
            imports: A list of GN import files.
        """
        for gni in imports:
            self.write_import(gni)

    def write_config(self, config: GnConfig) -> None:
        """Adds a GN config.

        Args:
            config: The GN config data to write.
        """
        if not config:
            return
        if not config.label:
            raise MalformedGnError('missing label for `config`')
        self.write_target_start('config', config.label.name())
        for flag in GN_CONFIG_FLAGS:
            self.write_list(flag, config.get(flag))
        self.write_end()

    def write_target(self, target: GnTarget) -> None:
        """Write a GN target.

        Args:
            target: The GN target data to write.
        """
        self.write_comment(
            f'Generated from //{target.package()}:{target.name()}'
        )
        self.write_target_start(target.type(), target.name())

        # GN use no `visibility` to indicate publicly visibile.
        scopes = filter(lambda s: str(s) != '//*', target.visibility)
        visibility = [target.make_relative(scope) for scope in scopes]
        self.write_list('visibility', visibility)

        if not target.check_includes:
            self.write('check_includes = false')
        self.write_list('public', [str(path) for path in target.public])
        self.write_list('sources', [str(path) for path in target.sources])
        self.write_list('inputs', [str(path) for path in target.inputs])

        for flag in GN_CONFIG_FLAGS:
            self.write_list(flag, target.config.get(flag))
        self._write_relative('public_configs', target, target.public_configs)
        self._write_relative('configs', target, target.configs)
        self._write_relative('remove_configs', target, target.remove_configs)

        self._write_relative('public_deps', target, target.public_deps)
        self._write_relative('deps', target, target.deps)
        self.write_end()

    def _write_relative(
        self, var_name: str, target: GnTarget, labels: Iterable[GnLabel]
    ) -> None:
        """Write a list of labels relative to a target.

        Args:
            var_name: The name of the GN list variable.
            target: The GN target to rebase the labels to.
            labels: The labels to write to the list.
        """
        self.write_list(var_name, self._resolve(target, labels))

    def _resolve(
        self, target: GnTarget, labels: Iterable[GnLabel]
    ) -> Iterator[str]:
        """Returns rewritten labels.

        If this label has a repo, it must be a key in this object's `repos` and
        will be replaced by the corresponding value. If this label  is a key in
        this object's `aliases`, it will be replaced by the corresponding value.

        Args:
            labels: The labels to resolve.
        """
        for label in labels:
            repo = label.repo()
            if repo:
                label.resolve_repo(self.repos[repo])
            label = GnLabel(self.aliases.get(str(label), str(label)))
            yield target.make_relative(label)

    def write_target_start(
        self, target_type: str, target_name: Optional[str] = None
    ) -> None:
        """Begins a GN target of the given type.

        Args:
            target_type: The type of the GN target.
            target_name: The name of the GN target.
        """
        if target_name:
            self.write(f'{target_type}("{target_name}") {{')
            self._indent(target_name)
        else:
            self.write(f'{target_type}() {{')
            self._indent(target_type)

    def write_list(
        self, var_name: str, items: Iterable[str], reorder: bool = True
    ) -> None:
        """Adds a named GN list of the given items, if non-empty.

        Args:
            var_name: The name of the GN list variable.
            items: The list items to write as strings.
            reorder: If true, the list is sorted lexicographically.
        """
        items = list(items)
        if not items:
            return
        self.write(f'{var_name} = [')
        self._indent(var_name)
        if reorder:
            items = sorted(items)
        for item in items:
            self.write(f'"{str(item)}",')
        self._outdent()
        self.write(']')

    def write_scope(self, var_name: str) -> None:
        """Begins a named GN scope.

        Args:
            var_name: The name of the GN scope variable.
        """
        self.write(f'{var_name} = {{')
        self._indent(var_name)

    def write_if(self, cond: str) -> None:
        """Begins a GN 'if' condition.

        Args:
            cond: The conditional expression.
        """
        self.write(f'if ({cond}) {{')
        self._indent(cond)

    def write_else_if(self, cond: str) -> None:
        """Adds another GN 'if' condition to a previous 'if' condition.

        Args:
            cond: The conditional expression.
        """
        self._outdent()
        self.write(f'}} else if ({cond}) {{')
        self._indent(cond)

    def write_else(self) -> None:
        """Adds a GN 'else' clause to a previous 'if' condition."""
        last = self._outdent()
        self.write('} else {')
        self._indent(f'!({last})')

    def write_end(self) -> None:
        """Ends a target, scope, or 'if' condition'."""
        self._outdent()
        self.write('}')
        self._needs_blank = True

    def write_blank(self) -> None:
        """Adds a blank line."""
        print('', file=self._file)
        self._needs_blank = False

    def write_preformatted(self, preformatted: str) -> None:
        """Adds text with minimal formatting.

        The only formatting applied to the given text is to strip any leading
        whitespace. This allows calls to be more readable by allowing
        preformatted text to start on a new line, e.g.

            _write_preformatted('''
          preformatted line 1
          preformatted line 2
          preformatted line 3''')

        Args:
            preformatted: The text to write.
        """
        print(preformatted.lstrip(), file=self._file)

    def write(self, text: str) -> None:
        """Writes to the file, appropriately indented.

        Args:
            text: The text to indent and write.
        """
        if self._needs_blank:
            self.write_blank()
        print(f'{self._margin}{text}', file=self._file)

    def _indent(self, scope: str) -> None:
        """Increases the current margin.

        Saves the scope of indent to aid in debugging. For example, trying to
        use incorrect code such as

        ```
          self.write_if('foo')
          self.write_comment('bar')
          self.write_else_if('baz')
        ```

        will throw an exception due to the missing `write_end`. The exception
        will note that 'baz' was opened but not closed.

        Args:
            scope: The name of the scope (for debugging).
        """
        self._scopes.append(scope)
        self._margin += '  '

    def _outdent(self) -> str:
        """Decreases the current margin."""
        if not self._scopes:
            raise MalformedGnError('scope closed unexpectedly')
        last = self._scopes.pop()
        self._margin = self._margin[2:]
        self._needs_blank = False
        return last

    def seal(self) -> None:
        """Instructs the object that no more writes will occur."""
        if self._scopes:
            raise MalformedGnError(f'unclosed scope(s): {self._scopes}')


def gn_format(gn_file: Path) -> None:
    """Calls `gn format` on a BUILD.gn or GN import file."""
    subprocess.check_call(['gn', 'format', gn_file])


class GnFile:
    """Represents an open BUILD.gn file that is formatted on close.

    Typical usage:

        with GnFile('/path/to/BUILD.gn', 'my-package') as build_gn:
          build_gn.write_...

    where "write_..." refers to any of the "write" methods of `GnWriter`.
    """

    def __init__(
        self, pathname: PurePath, package: Optional[str] = None
    ) -> None:
        if pathname.name != 'BUILD.gn' and pathname.suffix != '.gni':
            raise MalformedGnError(f'invalid GN filename: {pathname}')
        os.makedirs(pathname.parent, exist_ok=True)
        self._pathname: PurePath = pathname
        self._package: Optional[str] = package
        self._file: IO
        self._writer: GnWriter

    def __enter__(self) -> GnWriter:
        """Opens the GN file."""
        self._file = open(self._pathname, 'w+')
        self._writer = GnWriter(self._file)
        self._writer.write_preformatted(COPYRIGHT_HEADER)
        file = PurePath(*PurePath(__file__).parts[-2:])
        self._writer.write_comment(
            f'This file was automatically generated by {file}'
        )
        if self._package:
            self._writer.write_comment(
                f'It contains GN build targets for {self._package}.'
            )
        self._writer.write_blank()
        return self._writer

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Closes the GN file and formats it."""
        self._file.close()
        gn_format(Path(self._pathname))
